"""
..todo::
    add/remove events (check component indices)
    add/remove durations
    add/remove dependencies

- event times (especially considerung time modulo cycle time)
- event /parameter/ performer timelines
- section states (especially piecewise poly times)

Notes
-----
Since the EventHandler defines an interface, that requires the
implementation of some methods, a HandlerFixture class is defined.

Maybe this is too complicated, just use Process instead?
"""

import unittest

import CADETProcess
import numpy as np
from addict import Dict
from CADETProcess.dataStructure import (
    Float,
    NdPolynomial,
    Polynomial,
    SizedList,
    SizedNdArray,
    SizedTuple,
    Structure,
    Switch,
)

plot = True


class PerformerFixture(Structure):
    n_entries = 4
    n_coeff = 4

    scalar_float = Float(default=0)
    switch = Switch(valid=[-1, 1], default=1)
    sized_tuple = SizedTuple(size=2, default=(1, 1))
    array_1d = SizedList(size=4, default=0)
    array_1d_single = SizedList(size=1, default=0)
    ndarray = SizedNdArray(size=(2, 4), default=0)
    ndarray_no_default = SizedNdArray(size=(2, 4))
    array_1d_poly = Polynomial(n_coeff=4, default=0)
    ndarray_poly = NdPolynomial(n_entries=2, n_coeff=4, default=0)
    ndarray_poly_dep = NdPolynomial(size=("n_entries", "n_coeff"), default=0)

    _parameters = [
        "scalar_float",
        "sized_tuple",
        "switch",
        "array_1d",
        "array_1d_single",
        "ndarray",
        "ndarray_no_default",
        "array_1d_poly",
        "ndarray_poly",
        "ndarray_poly_dep",
    ]
    _section_dependent_parameters = [
        "scalar_float",
        "sized_tuple",
        "array_1d",
        "array_1d_single",
        "ndarray",
        "ndarray_no_default",
        "array_1d_poly",
        "ndarray_poly",
        "ndarray_poly_dep",
    ]

    @property
    def section_dependent_parameters(self):
        parameters = {
            param: getattr(self, param) for param in self._section_dependent_parameters
        }
        return parameters


class HandlerFixture(CADETProcess.dynamicEvents.EventHandler):
    def __init__(self):
        self.name = None
        self.performer = PerformerFixture()
        super().__init__()

    @property
    def parameters(self):
        parameters = super().parameters

        parameters["performer"] = self.performer.parameters

        return Dict(parameters)

    @parameters.setter
    def parameters(self, parameters):
        try:
            self.performer.parameters = parameters.pop("performer")
        except KeyError:
            pass

        super(HandlerFixture, self.__class__).parameters.fset(self, parameters)

    @property
    def section_dependent_parameters(self):
        parameters = Dict()
        parameters["performer"] = self.performer.section_dependent_parameters
        return parameters

    @property
    def polynomial_parameters(self):
        parameters = Dict()
        parameters["performer"] = self.performer.polynomial_parameters
        return parameters


class Test_Events(unittest.TestCase):
    def setup_event_handler(self, add_events=False):
        event_handler = HandlerFixture()
        event_handler.cycle_time = 20

        if add_events:
            evt = event_handler.add_event("evt0", "performer.scalar_float", 0)
            evt = event_handler.add_event("evt1", "performer.scalar_float", 1)
            evt = event_handler.add_event("evt2", "performer.sized_tuple", (2, 1))
            evt = event_handler.add_event("evt3", "performer.sized_tuple", (3, 2))
            evt = event_handler.add_event("evt4", "performer.sized_tuple", (3, 3))

        return event_handler

    def setUp(self):
        self.event_handler = self.setup_event_handler()

    def test_exceptions(self):
        event_handler = self.setup_event_handler()

        # Invalid path
        with self.assertRaises(CADETProcess.CADETProcessError):
            event_handler.add_event("wrong_path", "performer.wrong", 1)

        # Invalid state
        with self.assertRaises(TypeError):
            event_handler.add_event("wrong_value", "performer.scalar_float", "wrong")

        # Duplicate name
        with self.assertRaises(CADETProcess.CADETProcessError):
            event_handler.add_event("duplicate", "performer.scalar_float", 1)
            event_handler.add_event("duplicate", "performer.scalar_float", 1)

        # Not section dependent
        with self.assertRaises(CADETProcess.CADETProcessError):
            event_handler.add_event("not_sec_dependent", "performer.switch", 1)

    def test_event_scalar(self):
        event_handler = self.event_handler

        evt = event_handler.add_event("trivial", "performer.scalar_float", 1, time=0)
        self.assertEqual(evt.state, 1)
        self.assertEqual(evt.full_state, 1)
        self.assertEqual(evt.n_entries, 1)
        self.assertEqual(evt.indices, None)
        self.assertEqual(evt.full_indices, [])
        self.assertEqual(evt.n_indices, 0)
        self.assertEqual(event_handler.performer.scalar_float, 1)

        # Raise Error for indices
        with self.assertRaises(IndexError):
            evt = event_handler.add_event(
                "param_has_no_indices", "performer.scalar_float", 1, time=0, indices=1
            )

    def test_event_array_1d_single(self):
        """
        Extra test for array with single entry.

        See also: https://github.com/fau-advanced-separations/CADET-Process/pull/68
        """
        event_handler = self.event_handler

        # No index
        evt = event_handler.add_event("trivial", "performer.array_1d_single", 1, time=0)
        self.assertEqual(evt.state, 1)
        self.assertEqual(evt.full_state, [1])
        self.assertEqual(evt.n_entries, 1)
        self.assertEqual(evt.indices, [(slice(None, None, None),)])
        self.assertEqual(evt.full_indices, [(0,)])
        self.assertEqual(evt.n_indices, 1)
        self.assertEqual(event_handler.performer.array_1d_single, [1])

        # Explicit Index
        evt = event_handler.add_event(
            "trivial_1", "performer.array_1d_single", 2, time=0, indices=0
        )
        self.assertEqual(evt.state, 2)
        self.assertEqual(evt.full_state, [2])
        self.assertEqual(evt.n_entries, 1)
        self.assertEqual(evt.indices, [(0,)])
        self.assertEqual(evt.full_indices, [(0,)])
        self.assertEqual(evt.n_indices, 1)
        self.assertEqual(event_handler.performer.array_1d_single, [2])

        # Raise Error for exceeding indices
        with self.assertRaises(IndexError):
            evt = event_handler.add_event(
                "param_has_no_indices",
                "performer.array_1d_single",
                1,
                time=0,
                indices=1,
            )

    def test_event_1D(self):
        # TODO: Check too many / few indices
        event_handler = self.event_handler

        # Add event for single entry in 1D list / array
        event_handler.performer.array_1d
        evt = event_handler.add_event(
            "1D_single", "performer.array_1d", 1, indices=0, time=0
        )
        self.assertEqual(evt.state, 1)
        self.assertEqual(evt.full_state, [1])
        self.assertEqual(evt.n_entries, 1)
        self.assertEqual(evt.indices, [(0,)])
        self.assertEqual(evt.full_indices, [(0,)])
        self.assertEqual(evt.n_indices, 1)
        np.testing.assert_equal(event_handler.performer.array_1d, [1, 0, 0, 0])

        # Add event for multiple entries in 1D list / array
        # TODO: When creating section, missing parameters must be read at that time!
        evt = event_handler.add_event(
            "1D_multi", "performer.array_1d", [0, 1], indices=[0, 1], time=1
        )
        np.testing.assert_equal(evt.state, [0, 1])
        self.assertEqual(evt.full_state, [0, 1])
        self.assertEqual(evt.n_entries, 2)
        np.testing.assert_equal(evt.indices, [(0,), (1,)])
        self.assertEqual(evt.full_indices, [(0,), (1,)])
        self.assertEqual(evt.n_indices, 2)
        np.testing.assert_equal(event_handler.performer.array_1d, [0, 1, 0, 0])

        # Add event for multiple entries in 1D list / array with custom order
        evt = event_handler.add_event(
            "1D_multi_order", "performer.array_1d", [2, 3], indices=[1, 2], time=2
        )
        np.testing.assert_equal(evt.state, [2, 3])
        np.testing.assert_equal(evt.full_state, [2, 3])
        self.assertEqual(evt.n_entries, 2)
        np.testing.assert_equal(evt.indices, [(1,), (2,)])
        np.testing.assert_equal(evt.full_indices, [(1,), (2,)])
        self.assertEqual(evt.n_indices, 2)
        np.testing.assert_equal(event_handler.performer.array_1d, [0, 2, 3, 0])

        # Add event for all entries in 1D list / array
        evt = event_handler.add_event(
            "1D_all", "performer.array_1d", [0, 0, 1], indices=[0, 1, 2], time=3
        )
        np.testing.assert_equal(evt.state, [0, 0, 1])
        np.testing.assert_equal(evt.full_state, [0, 0, 1])
        self.assertEqual(evt.n_entries, 3)
        np.testing.assert_equal(evt.indices, [(0,), (1,), (2,)])
        np.testing.assert_equal(evt.full_indices, [(0,), (1,), (2,)])
        self.assertEqual(evt.n_indices, 3)
        np.testing.assert_equal(event_handler.performer.array_1d, [0, 0, 1, 0])

        # Set all values without indices
        evt = event_handler.add_event(
            "1D_all_slice_all_no_indices", "performer.array_1d", [0, 1, 2, 3], time=4
        )
        np.testing.assert_equal(evt.state, [0, 1, 2, 3])
        np.testing.assert_equal(evt.full_state, [0, 1, 2, 3])
        self.assertEqual(evt.n_entries, 4)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(evt.full_indices, [(0,), (1,), (2,), (3,)])
        self.assertEqual(evt.n_indices, 4)
        np.testing.assert_equal(event_handler.performer.array_1d, [0, 1, 2, 3])

        # Set all values using slicing notation.
        evt = event_handler.add_event(
            "1D_all_slice", "performer.array_1d", [2, 3, 4, 5], indices=np.s_[:], time=5
        )
        np.testing.assert_equal(evt.state, [2, 3, 4, 5])
        np.testing.assert_equal(evt.full_state, [2, 3, 4, 5])
        self.assertEqual(evt.n_entries, 4)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(evt.full_indices, [(0,), (1,), (2,), (3,)])
        self.assertEqual(evt.n_indices, 4)
        np.testing.assert_equal(event_handler.performer.array_1d, [2, 3, 4, 5])

        # Set all values to one value using slicing notation
        evt = event_handler.add_event(
            "1D_all_slice_one_value", "performer.array_1d", 1, indices=np.s_[:], time=6
        )
        np.testing.assert_equal(evt.state, 1)
        np.testing.assert_equal(evt.full_state, [1, 1, 1, 1])
        self.assertEqual(evt.n_entries, 4)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(evt.full_indices, [(0,), (1,), (2,), (3,)])
        self.assertEqual(evt.n_indices, 4)
        np.testing.assert_equal(event_handler.performer.array_1d, [1, 1, 1, 1])

        # Set all values from starting_index to one value using slice
        evt = event_handler.add_event(
            "1D_partial_slice", "performer.array_1d", 2, indices=np.s_[1:], time=7
        )
        np.testing.assert_equal(evt.state, 2)
        np.testing.assert_equal(evt.full_state, [2, 2, 2])
        self.assertEqual(evt.n_entries, 3)
        np.testing.assert_equal(evt.indices, [(slice(1, None, None),)])
        np.testing.assert_equal(evt.full_indices, [(1,), (2,), (3,)])
        self.assertEqual(evt.n_indices, 3)
        np.testing.assert_equal(event_handler.performer.array_1d, [1, 2, 2, 2])

        if plot:
            event_handler.plot_events()

        # Number of indices
        with self.assertRaises(ValueError):
            event_handler.add_event(
                "not_enough_entries", "performer.array_1d", [1, 2], indices=[2]
            )
        with self.assertRaises(ValueError):
            event_handler.add_event(
                "too_many_indices", "performer.array_1d", [1, 2], indices=[1, 2, 2]
            )

        # Index exceeds shape
        with self.assertRaises(IndexError):
            event_handler.add_event(
                "index_exceeds_shape", "performer.array_1d", [1, 2], indices=[1, 4]
            )

        # Duplicate entries for indices
        with self.assertRaises(ValueError):
            event_handler.add_event(
                "duplicate_entries", "performer.array_1d", [1, 2], indices=[1, 1]
            )

    def test_event_ndarray(self):
        event_handler = self.event_handler

        t = 0

        # Add event for single entry in ndarray
        t += 1
        evt = event_handler.add_event(
            "ndarray_single", "performer.ndarray", 1, indices=(0, 0), time=t
        )
        np.testing.assert_equal(evt.state, 1)
        self.assertEqual(evt.n_entries, 1)
        np.testing.assert_equal(evt.indices, [(0, 0)])
        np.testing.assert_equal(evt.full_indices, [(0, 0)])
        self.assertEqual(evt.n_indices, 1)
        np.testing.assert_equal(
            event_handler.performer.ndarray, [[1, 0, 0, 0], [0, 0, 0, 0]]
        )

        # Add event for multiple entries in array
        t += 1
        evt = event_handler.add_event(
            "ndarray_multiple",
            "performer.ndarray",
            [0, 1],
            indices=[(0, 0), (0, 1)],
            time=t,
        )
        np.testing.assert_equal(evt.state, [0, 1])
        np.testing.assert_equal(evt.full_state, [0, 1])
        self.assertEqual(evt.n_entries, 2)
        np.testing.assert_equal(evt.indices, [(0, 0), (0, 1)])
        np.testing.assert_equal(evt.full_indices, [(0, 0), (0, 1)])
        self.assertEqual(evt.n_indices, 2)
        np.testing.assert_equal(
            event_handler.performer.ndarray, [[0, 1, 0, 0], [0, 0, 0, 0]]
        )

        # Add event for multiple entries in array with custom order
        t += 1
        evt = event_handler.add_event(
            "ndarray_multiple_order",
            "performer.ndarray",
            [1, 0],
            indices=[(1, 0), (0, 1)],
            time=t,
        )
        np.testing.assert_equal(evt.state, [1, 0])
        np.testing.assert_equal(evt.full_state, [1, 0])
        self.assertEqual(evt.n_entries, 2)
        np.testing.assert_equal(evt.indices, [(1, 0), (0, 1)])
        np.testing.assert_equal(evt.full_indices, [(1, 0), (0, 1)])
        self.assertEqual(evt.n_indices, 2)
        np.testing.assert_equal(
            event_handler.performer.ndarray, [[0, 0, 0, 0], [1, 0, 0, 0]]
        )

        # Add event for all entries in array
        t += 1
        evt = event_handler.add_event(
            "ndarray_all",
            "performer.ndarray",
            [0, 1, 2, 3, 4, 5, 6, 7],
            indices=[(0, 0), (0, 1), (0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)],
            time=t,
        )
        np.testing.assert_equal(evt.state, [0, 1, 2, 3, 4, 5, 6, 7])
        self.assertEqual(evt.n_entries, 8)
        np.testing.assert_equal(
            evt.indices,
            [(0, 0), (0, 1), (0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)],
        )
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 0), (0, 1), (0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)],
        )
        self.assertEqual(evt.n_indices, 8)
        np.testing.assert_equal(
            event_handler.performer.ndarray, [[0, 1, 2, 3], [4, 5, 6, 7]]
        )

        # Set all values without indices
        t += 1
        evt = event_handler.add_event(
            "ndarray_all_no_indices",
            "performer.ndarray",
            [[8, 7, 6, 5], [4, 3, 2, 1]],
            time=t,
        )
        np.testing.assert_equal(evt.state, [[8, 7, 6, 5], [4, 3, 2, 1]])
        np.testing.assert_equal(evt.full_state, [8, 7, 6, 5, 4, 3, 2, 1])
        self.assertEqual(evt.n_entries, 8)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 0), (0, 1), (0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)],
        )
        self.assertEqual(evt.n_indices, 8)
        np.testing.assert_equal(
            event_handler.performer.ndarray, [[8, 7, 6, 5], [4, 3, 2, 1]]
        )

        # Set all values using slicing notation.
        t += 1
        evt = event_handler.add_event(
            "ndarray_all_slice",
            "performer.ndarray",
            [[0, 0, 0, 0], [1, 1, 1, 1]],
            indices=np.s_[:],
            time=t,
        )
        np.testing.assert_equal(evt.state, [[0, 0, 0, 0], [1, 1, 1, 1]])
        np.testing.assert_equal(evt.full_state, [0, 0, 0, 0, 1, 1, 1, 1])
        self.assertEqual(evt.n_entries, 8)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 0), (0, 1), (0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)],
        )
        self.assertEqual(evt.n_indices, 8)
        np.testing.assert_equal(
            event_handler.performer.ndarray, [[0, 0, 0, 0], [1, 1, 1, 1]]
        )

        # Set all values to one value using slicing notation
        t += 1
        evt = event_handler.add_event(
            "ndarray_all_slice_one_value",
            "performer.ndarray",
            1,
            indices=np.s_[:],
            time=t,
        )
        np.testing.assert_equal(evt.state, 1)
        np.testing.assert_equal(evt.full_state, [1, 1, 1, 1, 1, 1, 1, 1])
        self.assertEqual(evt.n_entries, 8)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 0), (0, 1), (0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)],
        )
        self.assertEqual(evt.n_indices, 8)
        np.testing.assert_equal(
            event_handler.performer.ndarray, [[1, 1, 1, 1], [1, 1, 1, 1]]
        )

        # TODO: Check slicing for rows/columns
        # Set all values of one row using slicing notation
        t += 1
        evt = event_handler.add_event(
            "ndarray_row_slice",
            "performer.ndarray",
            [0, 0, 0, 0],
            indices=np.s_[0, :],
            time=t,
        )
        np.testing.assert_equal(evt.state, [0, 0, 0, 0])
        np.testing.assert_equal(evt.full_state, [0, 0, 0, 0])
        self.assertEqual(evt.n_entries, 4)
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 0), (0, 1), (0, 2), (0, 3)],
        )
        self.assertEqual(evt.n_indices, 4)
        np.testing.assert_equal(
            event_handler.performer.ndarray, [[0, 0, 0, 0], [1, 1, 1, 1]]
        )

        # Set all values of one row to one value using slicing notation
        t += 1
        evt = event_handler.add_event(
            "ndarray_row_slice_one_value",
            "performer.ndarray",
            2,
            indices=np.s_[0, :],
            time=t,
        )
        np.testing.assert_equal(evt.state, 2)
        np.testing.assert_equal(evt.full_state, [2, 2, 2, 2])
        self.assertEqual(evt.n_entries, 4)
        np.testing.assert_equal(
            evt.indices,
            [(0, slice(None, None, None))],
        )
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 0), (0, 1), (0, 2), (0, 3)],
        )
        self.assertEqual(evt.n_indices, 4)
        np.testing.assert_equal(
            event_handler.performer.ndarray,
            [[2, 2, 2, 2], [1, 1, 1, 1]],
        )

        # Set all values of one column from starting_index to one value using slice
        t += 1
        evt = event_handler.add_event(
            "ndarray_column_partial_slice",
            "performer.ndarray",
            3,
            indices=np.s_[1:, 1],
            time=t,
        )
        np.testing.assert_equal(evt.state, 3)
        np.testing.assert_equal(evt.full_state, [3])
        self.assertEqual(evt.n_entries, 1)
        np.testing.assert_equal(evt.indices, [(slice(1, None, None), 1)])
        np.testing.assert_equal(evt.full_indices, [(1, 1)])
        self.assertEqual(evt.n_indices, 1)
        np.testing.assert_equal(
            event_handler.performer.ndarray,
            [[2, 2, 2, 2], [1, 3, 1, 1]],
        )

        # Check multiple slices
        t += 1
        evt = event_handler.add_event(
            "ndarray_multiple_slices",
            "performer.ndarray",
            [4, 5],
            indices=[np.s_[0, 1:], np.s_[1, :]],
            time=t,
        )
        np.testing.assert_equal(evt.state, [4, 5])
        np.testing.assert_equal(evt.full_state, [4, 4, 4, 5, 5, 5, 5])
        self.assertEqual(evt.n_entries, 7)
        np.testing.assert_equal(
            evt.indices,
            [(0, slice(1, None, None)), (1, slice(None, None, None))],
        )
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 1), (0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)],
        )
        self.assertEqual(evt.n_indices, 7)
        np.testing.assert_equal(
            event_handler.performer.ndarray,
            [[2, 4, 4, 4], [5, 5, 5, 5]],
        )

        if plot:
            event_handler.plot_events()

    def test_event_polynomial_1D(self):
        event_handler = self.event_handler

        # Add event relying on filling missing coefficients
        evt = event_handler.add_event(
            "1D_single_fill", "performer.array_1d_poly", 1, time=0
        )
        np.testing.assert_equal(evt.state, 1)
        np.testing.assert_equal(evt.full_state, [1, 0, 0, 0])
        self.assertEqual(evt.n_entries, 4)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(evt.full_indices, [(0,), (1,), (2,), (3,)])
        self.assertEqual(evt.n_indices, 4)
        np.testing.assert_equal(
            event_handler.performer.array_1d_poly,
            [1, 0, 0, 0],
        )

        # Fill all values to specific value
        evt = event_handler.add_event(
            "1D_multi_no_fill", "performer.array_1d_poly", 2, indices=np.s_[:], time=1
        )
        np.testing.assert_equal(evt.state, 2)
        np.testing.assert_equal(evt.full_state, [2, 2, 2, 2])
        self.assertEqual(evt.n_entries, 4)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(evt.full_indices, [(0,), (1,), (2,), (3,)])
        self.assertEqual(evt.n_indices, 4)
        np.testing.assert_equal(event_handler.performer.array_1d_poly, [2, 2, 2, 2])

        # Add event relying on filling missing coefficients
        evt = event_handler.add_event(
            "1D_multi_fill", "performer.array_1d_poly", [3, 3], time=2
        )
        np.testing.assert_equal(evt.state, [3, 3])
        np.testing.assert_equal(evt.full_state, [3, 3, 0, 0])
        self.assertEqual(evt.n_entries, 4)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(evt.full_indices, [(0,), (1,), (2,), (3,)])
        self.assertEqual(evt.n_indices, 4)
        np.testing.assert_equal(event_handler.performer.array_1d_poly, [3, 3, 0, 0])

        # Add event for single entry in array
        evt = event_handler.add_event(
            "1D_single", "performer.array_1d_poly", 1, indices=0, time=3
        )
        np.testing.assert_equal(evt.state, 1)
        np.testing.assert_equal(evt.full_state, [1])
        self.assertEqual(evt.n_entries, 1)
        np.testing.assert_equal(evt.indices, [(0,)])
        np.testing.assert_equal(evt.full_indices, [(0,)])
        self.assertEqual(evt.n_indices, 1)
        np.testing.assert_equal(event_handler.performer.array_1d_poly, [1, 3, 0, 0])

        # Add event for multiple entry in array
        evt = event_handler.add_event(
            "1D_multi", "performer.array_1d_poly", [2, 2], indices=[0, 1], time=4
        )
        np.testing.assert_equal(evt.state, [2, 2])
        np.testing.assert_equal(evt.full_state, [2, 2])
        self.assertEqual(evt.n_entries, 2)
        np.testing.assert_equal(evt.indices, [(0,), (1,)])
        np.testing.assert_equal(evt.full_indices, [(0,), (1,)])
        self.assertEqual(evt.n_indices, 2)
        np.testing.assert_equal(event_handler.performer.array_1d_poly, [2, 2, 0, 0])

        # Add event for single entry in array to check if list notation works
        evt = event_handler.add_event(
            "1D_multi_flat", "performer.array_1d_poly", [0], indices=[1], time=5
        )
        np.testing.assert_equal(evt.state, [0])
        np.testing.assert_equal(evt.full_state, [0])
        self.assertEqual(evt.n_entries, 1)
        np.testing.assert_equal(evt.indices, [(1,)])
        np.testing.assert_equal(evt.full_indices, [(1,)])
        self.assertEqual(evt.n_indices, 1)
        np.testing.assert_equal(event_handler.performer.array_1d_poly, [2, 0, 0, 0])

        if plot:
            event_handler.plot_events()

    def test_event_polynomial_ndarray(self):
        event_handler = self.event_handler

        t = 0

        # Add event relying for individual coefficient
        t += 1
        evt = event_handler.add_event(
            "single_coeff", "performer.ndarray_poly", 1, indices=(0, 1), time=t
        )
        np.testing.assert_equal(evt.state, 1)
        np.testing.assert_equal(evt.full_state, [1])
        self.assertEqual(evt.n_entries, 1)
        np.testing.assert_equal(evt.indices, [(0, 1)])
        np.testing.assert_equal(evt.full_indices, [(0, 1)])
        self.assertEqual(evt.n_indices, 1)
        np.testing.assert_equal(
            event_handler.performer.ndarray_poly,
            [
                [0, 1, 0, 0],
                [0, 0, 0, 0],
            ],
        )

        # Add event for multiple coefficients / indices
        t += 1
        evt = event_handler.add_event(
            "multi_coeffs",
            "performer.ndarray_poly",
            [1, 2],
            indices=[(0, 0), (1, 1)],
            time=t,
        )
        np.testing.assert_equal(evt.state, [1, 2])
        np.testing.assert_equal(evt.full_state, [1, 2])
        self.assertEqual(evt.n_entries, 2)
        np.testing.assert_equal(evt.indices, [(0, 0), (1, 1)])
        np.testing.assert_equal(evt.full_indices, [(0, 0), (1, 1)])
        self.assertEqual(evt.n_indices, 2)
        np.testing.assert_equal(
            event_handler.performer.ndarray_poly,
            [
                [1, 1, 0, 0],
                [0, 2, 0, 0],
            ],
        )

        # Add event relying on filling missing coefficients
        t += 1
        evt = event_handler.add_event("fill_all", "performer.ndarray_poly", 1, time=t)
        np.testing.assert_equal(evt.state, 1)
        np.testing.assert_equal(
            evt.full_state,
            [1, 0, 0, 0, 1, 0, 0, 0],
        )
        self.assertEqual(evt.n_entries, 8)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 0), (0, 1), (0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)],
        )
        self.assertEqual(evt.n_indices, 8)
        np.testing.assert_equal(
            event_handler.performer.ndarray_poly,
            [
                [1, 0, 0, 0],
                [1, 0, 0, 0],
            ],
        )
        # Add event relying on filling missing coefficients for single entry
        t += 1
        evt = event_handler.add_event(
            "multi_fill_single", "performer.ndarray_poly", 3, indices=0, time=t
        )
        np.testing.assert_equal(evt.state, 3)
        np.testing.assert_equal(
            evt.full_state,
            [3, 0, 0, 0],
        )
        self.assertEqual(evt.n_entries, 4)
        np.testing.assert_equal(evt.indices, [(0,)])
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 0), (0, 1), (0, 2), (0, 3)],
        )
        self.assertEqual(evt.n_indices, 4)
        np.testing.assert_equal(
            event_handler.performer.ndarray_poly,
            [
                [3, 0, 0, 0],
                [1, 0, 0, 0],
            ],
        )

        # Add event relying on filling missing coefficients with inhomogeneous shape
        t += 1
        evt = event_handler.add_event(
            "multi_fill_inhomogeneous", "performer.ndarray_poly", [2, [0, 1]], time=t
        )
        np.testing.assert_equal(evt.state, [2, [0, 1]])
        np.testing.assert_equal(
            evt.full_state,
            [2, 0, 0, 0, 0, 1, 0, 0],
        )
        self.assertEqual(evt.n_entries, 8)
        np.testing.assert_equal(evt.indices, [(slice(None, None, None),)])
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 0), (0, 1), (0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)],
        )
        self.assertEqual(evt.n_indices, 8)
        np.testing.assert_equal(
            event_handler.performer.ndarray_poly,
            [
                [2, 0, 0, 0],
                [0, 1, 0, 0],
            ],
        )

        # Add event using slicing (should work just like a regular ndarray)
        t += 1
        evt = event_handler.add_event(
            "multi_slice", "performer.ndarray_poly", [1], indices=[np.s_[0, :]], time=t
        )

        np.testing.assert_equal(evt.state, [1])
        np.testing.assert_equal(evt.full_state, [1, 1, 1, 1])
        self.assertEqual(evt.n_entries, 4)
        np.testing.assert_equal(evt.indices, [(0, slice(None, None, None))])
        np.testing.assert_equal(
            evt.full_indices,
            [(0, 0), (0, 1), (0, 2), (0, 3)],
        )
        self.assertEqual(evt.n_indices, 4)
        np.testing.assert_equal(
            event_handler.performer.ndarray_poly,
            [
                [1, 1, 1, 1],
                [0, 1, 0, 0],
            ],
        )

        if plot:
            event_handler.plot_events()

    def test_lwe(self):
        """This reproduces the situation from the LWE / SMA model in the examples."""
        event_handler = self.setup_event_handler()
        event_handler.name = "LWE"

        cycle_time = 110 * 60
        event_handler.cycle_time = cycle_time

        t_flush = 20 * 60

        t_gradient_start = 90
        gradient_duration = cycle_time - t_gradient_start - t_flush

        c_load = np.array([5, 1, 1, 1])
        c_wash = np.array([5, 0, 0, 0])
        c_elute = np.array([1000, 0, 0, 0])
        gradient_slope = (c_elute - c_wash) / gradient_duration
        c_gradient_poly = np.array(list(zip(c_wash, gradient_slope)))
        c_gradient_poly[0][1] = 0.4
        c_gradient_poly = c_gradient_poly.tolist()

        event_handler.add_event("load", "performer.ndarray_poly_dep", c_load, time=0)
        event_handler.add_event(
            "wash", "performer.ndarray_poly_dep", c_wash, time=t_flush
        )
        event_handler.add_event(
            "grad_start",
            "performer.ndarray_poly_dep",
            c_gradient_poly,
            time=t_gradient_start,
        )
        event_handler.add_event(
            "Flush",
            "performer.ndarray_poly_dep",
            c_elute,
            time=t_gradient_start + gradient_duration,
        )

        if plot:
            event_handler.plot_events()

    def test_event_times(self):
        event_handler = self.setup_event_handler(add_events=True)

        self.assertEqual(event_handler.event_times, [0])

        event_handler.evt0.time = 1
        self.assertEqual(event_handler.event_times, [0, 1])

        event_handler.cycle_time = 10
        event_handler.evt0.time = 11
        self.assertEqual(event_handler.event_times, [0, 1])

    def test_dependencies(self):
        event_handler = self.setup_event_handler(add_events=True)

        event_handler.add_event_dependency("evt1", "evt0")
        self.assertEqual(event_handler.event_times, [0])

        event_handler.evt0.time = 1
        self.assertEqual(event_handler.event_times, [0, 1])

        event_handler.add_event_dependency("evt2", "evt1", 2)
        self.assertEqual(event_handler.event_times, [0, 1, 2])

        event_handler.add_event_dependency("evt3", ["evt1", "evt0"], [2, 1])
        self.assertEqual(event_handler.event_times, [0, 1, 2, 3])

        # Dependent event
        with self.assertRaises(CADETProcess.CADETProcessError):
            event_handler.evt1.time = 1

        # Event does not exist
        with self.assertRaises(CADETProcess.CADETProcessError):
            event_handler.add_event_dependency("evt3", "evt0")

        # Duplicate dependency
        with self.assertRaises(CADETProcess.CADETProcessError):
            event_handler.add_event_dependency("evt1", "evt0")

        # Linear factors not matching
        with self.assertRaises(CADETProcess.CADETProcessError):
            event_handler.add_event_dependency("evt1", "evt0", [1, 1])

    def test_section_states(self):
        pass

    def test_timelines(self):
        pass

    def test_duplicate_event_times(self):
        event_handler = self.setup_event_handler()

        event_handler.add_event("evt0", "performer.scalar_float", 0, 0)
        event_handler.add_event("evt1", "performer.scalar_float", 1, 1)

        event_handler.check_config()

        event_handler.evt1.time = 0

        self.assertFalse(event_handler.check_duplicate_events())

    def test_uninitialized_indices(self):
        event_handler = self.setup_event_handler()

        event_handler.add_event(
            "evt0", "performer.ndarray_no_default", 0, indices=(0, 0)
        )

        self.assertFalse(event_handler.check_uninitialized_indices())


if __name__ == "__main__":
    plot = True
    import matplotlib.pyplot as plt

    plt.close("all")
    unittest.main()



import platform

import psutil

uname = platform.uname()
cpu_freq = psutil.cpu_freq()
memory = psutil.virtual_memory()
memory_total = psutil._common.bytes2human(memory.total)

system_information = {
    "system": uname.system,
    "release": uname.release,
    "machine": uname.machine,
    "processor": uname.processor,
    "n_cores": psutil.cpu_count(logical=True),
    "n_cores_physical": psutil.cpu_count(logical=False),
    "max_frequency": cpu_freq.max,
    "min_frequency": cpu_freq.min,
    "memory_total": memory_total,
}



import unittest

import numpy as np
from CADETProcess import CADETProcessError
from CADETProcess.optimization import Individual, ParetoFront, Population

from tests.test_individual import setup_individual

enable_plot = False


def setup_population(n_ind, n_vars, n_obj, n_nonlin=0, n_meta=0, rng=None):
    population = Population()

    if rng is None:
        rng = np.random.default_rng(12345)

    for i in range(n_ind):
        ind = setup_individual(n_vars, n_obj, n_nonlin, n_meta, rng)
        population.add_individual(ind)

    return population


class TestPopulation(unittest.TestCase):
    def setUp(self):
        x = [1, 2]
        f = [-1]
        self.individual_1 = Individual(x, f=f)

        x = [2, 3]
        f = [-2]
        self.individual_2 = Individual(x, f=f)

        x = [1.001, 2]
        f = [-1.001]
        self.individual_similar = Individual(x, f=f)

        x = [1, 2]
        f = [-1, -2]
        self.individual_multi_1 = Individual(x, f=f)

        x = [1.001, 2]
        f = [-1.001, -2]
        self.individual_multi_2 = Individual(x, f=f)

        x = [1, 2]
        f = [-1]
        g = [3]
        self.individual_constr_1 = Individual(x, f=f, g=g)

        x = [2, 3]
        f = [-2]
        g = [0]
        self.individual_constr_2 = Individual(x, f=f, g=g)

        x = [2, 3]
        f = [-2, -2]
        g = [0]
        m = [-4]
        self.individual_multi_meta_1 = Individual(x, f=f, m=m)

        self.population = Population()
        self.population.add_individual(self.individual_1)
        self.population.add_individual(self.individual_2)
        self.population.add_individual(self.individual_similar)

        self.population_multi = Population()
        self.population_multi.add_individual(self.individual_multi_1)
        self.population_multi.add_individual(self.individual_multi_2)

        self.population_constr = Population()
        self.population_constr.add_individual(self.individual_constr_1)
        self.population_constr.add_individual(self.individual_constr_2)

        self.population_meta = Population()
        self.population_meta.add_individual(self.individual_multi_meta_1)

    def test_dimensions(self):
        dimensions_expected = (2, 1, 0, 0)
        dimensions = self.population.dimensions
        self.assertEqual(dimensions, dimensions_expected)

        self.assertEqual(2, self.population.n_x)
        self.assertEqual(1, self.population.n_f)
        self.assertEqual(0, self.population.n_g)
        self.assertEqual(0, self.population.n_m)

        dimensions_expected = (2, 1, 1, 0)
        dimensions = self.population_constr.dimensions
        self.assertEqual(dimensions, dimensions_expected)

        dimensions_expected = (2, 1, 1, 0)
        dimensions = self.population_constr.dimensions
        self.assertEqual(dimensions, dimensions_expected)

        self.assertEqual(2, self.population_meta.n_x)
        self.assertEqual(2, self.population_meta.n_f)
        self.assertEqual(0, self.population_meta.n_g)
        self.assertEqual(1, self.population_meta.n_m)

    def test_values(self):
        x_expected = np.array(
            [
                [1, 2],
                [2, 3],
                [1.001, 2],
            ]
        )
        x = self.population.x
        np.testing.assert_almost_equal(x, x_expected)

        f_expected = np.array(
            [
                [-1],
                [-2],
                [-1.001],
            ]
        )
        f = self.population.f
        np.testing.assert_almost_equal(f, f_expected)

        g_expected = np.array(
            [
                [3],
                [0],
            ]
        )
        g = self.population_constr.g
        np.testing.assert_almost_equal(g, g_expected)

    def test_add_remove(self):
        with self.assertRaises(TypeError):
            self.population.add_individual("foo")

        self.population.add_individual(self.individual_1, ignore_duplicate=True)

        with self.assertRaises(CADETProcessError):
            self.population.add_individual(self.individual_1, ignore_duplicate=False)

        new_individual = Individual([9, 10], f=[3], g=[-1])
        with self.assertRaises(CADETProcessError):
            self.population.add_individual(new_individual)

        self.population_constr.add_individual(new_individual)

        self.assertFalse(new_individual in self.population)
        self.assertTrue(self.individual_1 in self.population_constr)

        x_expected = np.array(
            [
                [1, 2],
                [2, 3],
                [9, 10],
            ]
        )
        x = self.population_constr.x
        np.testing.assert_almost_equal(x, x_expected)

        f_expected = np.array(
            [
                [-1],
                [-2],
                [3],
            ]
        )
        f = self.population_constr.f
        np.testing.assert_almost_equal(f, f_expected)

        g_expected = np.array(
            [
                [3],
                [0],
                [-1],
            ]
        )
        g = self.population_constr.g
        np.testing.assert_almost_equal(g, g_expected)

        with self.assertRaises(TypeError):
            self.population.remove_individual("foo")

        with self.assertRaises(CADETProcessError):
            self.population.remove_individual(new_individual)

        self.population_constr.remove_individual(new_individual)
        x_expected = np.array(
            [
                [1, 2],
                [2, 3],
            ]
        )
        x = self.population_constr.x
        np.testing.assert_almost_equal(x, x_expected)

        f_expected = np.array(
            [
                [-1],
                [-2],
            ]
        )
        f = self.population_constr.f
        np.testing.assert_almost_equal(f, f_expected)

        g_expected = np.array(
            [
                [3],
                [0],
            ]
        )
        g = self.population_constr.g
        np.testing.assert_almost_equal(g, g_expected)

    def test_min_max(self):
        f_min_expected = [-2]
        f_min = self.population_constr.f_min
        np.testing.assert_almost_equal(f_min, f_min_expected)

        f_max_expected = [-1]
        f_max = self.population_constr.f_max
        np.testing.assert_almost_equal(f_max, f_max_expected)

        g_min_expected = [0]
        g_min = self.population_constr.g_min
        np.testing.assert_almost_equal(g_min, g_min_expected)

        g_max_expected = [3]
        g_max = self.population_constr.g_max
        np.testing.assert_almost_equal(g_max, g_max_expected)

    def test_plot(self):
        if enable_plot:
            pass

    def test_to_dict(self):
        # Test that Population can be converted to a dictionary
        population_dict = self.population.to_dict()
        individuals_list = population_dict["individuals"]
        self.assertEqual(len(individuals_list), 3)
        self.assertEqual(population_dict["id"], str(self.population.id))

    def test_from_dict(self):
        # Test that a Population can be created from a dictionary
        population_dict = self.population.to_dict()
        new_population = Population.from_dict(population_dict)
        self.assertEqual(new_population.id, self.population.id)
        self.assertEqual(len(new_population), len(self.population))

        self.assertTrue(self.individual_1 in new_population)


class TestPareto(unittest.TestCase):
    def setUp(self):
        front = ParetoFront(3)

        population = [
            Individual([1, 2], [1, 2, 3]),
        ]

        front.update(population)

        population = [
            Individual([1, 2.01], [1, 2, 3.01]),
            Individual([4, 5], [4, 5, 6]),
        ]

        front.update(population)


if __name__ == "__main__":
    enable_plot = True

    unittest.main()



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% tags=["remove-cell"]
from pathlib import Path
import sys

root_dir = Path('../../../../').resolve()
sys.path.append(root_dir.as_posix())

# %% [markdown]
# (fit_column_transport)=
# # Fit Column Transport Parameters
# When characterizing a chromatographic system column, some parameters of the model can be directly measured (e.g. column length, diameter) or are provided by the resin manufacturer (e.g. particle radius).
# However, other transport parameters like axial dispersion and porosities need to be determined through experiments.
#
# One approach to determine these parameters is the inverse method.
# By adjusting the values of the parameters in the simulation model and comparing the resulting behavior to the experimental data, the optimal parameter values that match the observed behavior can be found.
#
# ## Fit Bed Porosity and Axial Dispersion
# The bed porosity is an important parameter that describes the fraction of the column volume that is available for fluid flow.
# Axial dispersion is a transport parameter that describes the spreading of a solute or tracer as it moves through a chromatographic system column along the length of the column.
# It occurs due to differences in velocity between different layers of the fluid, leading to radial mixing of the fluid and the solutes.
#
# To determine the value of these parameters, an experiment is conducted using a non-pore-penetrating tracer.
# The tracer is injected into the column and its concentration at the column outlet is measured and compared to the concentration predicted by simulation results.

# %%
import numpy as np
data = np.loadtxt('experimental_data/non_pore_penetrating_tracer.csv', delimiter=',')

time_experiment = data[:, 0]
c_experiment = data[:, 1]

from CADETProcess.reference import ReferenceIO
tracer_peak = ReferenceIO(
    'Tracer Peak', time_experiment, c_experiment
)

if __name__ == '__main__':
    _ = tracer_peak.plot()

# %% [markdown]
# ### Reference Model
# To accurately model a chromatographic system, it is crucial to establish a reference model that closely resembles the real system.
# The reference model allows for the determination of parameter influences and optimization of their values.
# Arbitrary values can be set for the unknown parameters such as `bed_porosity` and `axial_dispersion`, since they will be optimized.
# In order to model that the non-penetrating tracer does in fact not enter the pore, the `film_diffusion` needs to be set to $0$.
# Moreover, `particle_porosity` will be determined using a separate experiment.

# %%
from CADETProcess.processModel import ComponentSystem
component_system = ComponentSystem(['Non-penetrating Tracer'])

# %%
from CADETProcess.processModel import Inlet, Outlet, LumpedRateModelWithPores

feed = Inlet(component_system, name='feed')
feed.c = [0.0005]

eluent = Inlet(component_system, name='eluent')
eluent.c = [0]

column = LumpedRateModelWithPores(component_system, name='column')

column.length = 0.1
column.diameter = 0.0077
column.particle_radius = 34e-6

column.axial_dispersion = 1e-8
column.bed_porosity = 0.3
column.particle_porosity = 0.8
column.film_diffusion = [0]

outlet = Outlet(component_system, name='outlet')

# %%
from CADETProcess.processModel import FlowSheet

flow_sheet = FlowSheet(component_system)

flow_sheet.add_unit(feed)
flow_sheet.add_unit(eluent)
flow_sheet.add_unit(column)
flow_sheet.add_unit(outlet)

flow_sheet.add_connection(feed, column)
flow_sheet.add_connection(eluent, column)
flow_sheet.add_connection(column, outlet)

# %%
from CADETProcess.processModel import Process

Q_ml_min = 0.5  # ml/min
Q_m3_s = Q_ml_min / (60 * 1e6)
V_tracer = 50e-9  # m3

process = Process(flow_sheet, 'Tracer')
process.cycle_time = 15 * 60

process.add_event(
    'feed_on',
    'flow_sheet.feed.flow_rate',
    Q_m3_s, 0
)
process.add_event(
    'feed_off',
    'flow_sheet.feed.flow_rate',
    0,
    V_tracer / Q_m3_s
)

process.add_event(
    'feed_water_on',
    'flow_sheet.eluent.flow_rate',
     Q_m3_s,
     V_tracer / Q_m3_s
)

process.add_event(
    'eluent_off',
    'flow_sheet.eluent.flow_rate',
    0,
    process.cycle_time
)

# %% [markdown]
# ### Simulator

# %%
from CADETProcess.simulator import Cadet
simulator = Cadet()

if __name__ == '__main__':
    simulation_results = simulator.simulate(process)
    _ = simulation_results.solution.outlet.inlet.plot()

# %% [markdown]
# ### Comparator

# %%
from CADETProcess.comparison import Comparator

comparator = Comparator()
comparator.add_reference(tracer_peak)
comparator.add_difference_metric(
    'NRMSE', tracer_peak, 'outlet.outlet',
)

if __name__ == '__main__':
    comparator.plot_comparison(simulation_results)

# %% [markdown]
# ### Optimization Problem

# %%
from CADETProcess.optimization import OptimizationProblem
optimization_problem = OptimizationProblem('bed_porosity_axial_dispersion')

optimization_problem.add_evaluation_object(process)

optimization_problem.add_variable(
    name='bed_porosity', parameter_path='flow_sheet.column.bed_porosity',
    lb=0.1, ub=0.6,
    transform='auto'
)

optimization_problem.add_variable(
    name='axial_dispersion', parameter_path='flow_sheet.column.axial_dispersion',
    lb=1e-10, ub=0.1,
    transform='auto'
)

optimization_problem.add_evaluator(simulator)

optimization_problem.add_objective(
    comparator,
    n_objectives=comparator.n_metrics,
    requires=[simulator]
)


def callback(simulation_results, individual, evaluation_object, callbacks_dir='./'):
    comparator.plot_comparison(
        simulation_results,
        file_name=f'{callbacks_dir}/{individual.id}_{evaluation_object}_comparison.png',
        show=False
    )


optimization_problem.add_callback(callback, requires=[simulator])

# %% [markdown]
# ### Optimizer

# %%
from CADETProcess.optimization import U_NSGA3
optimizer = U_NSGA3()

# %% [markdown]
# ```{note}
# For performance reasons, the optimization is currently not run when building the documentation.
# In future, we will try to sideload pre-computed results to also discuss them here.
# ```
#
# ```
# if __name__ == '__main__':
#     optimization_results = optimizer.optimize(
#         optimization_problem,
#         use_checkpoint=True
#     )
# ```



"""
===============================================
ModelBuilder (:mod:`CADETProcess.modelBuilder`)
===============================================

.. currentmodule:: CADETProcess.modelBuilder

The ``modelBuilder`` module provides functionality for setting up complex ``Process``
models

CarouselBuilder
===============

A module for building carousel systems like SMB, MCSGP, etc.

.. autosummary::
   :toctree: generated/

   SerialZone
   ParallelZone
   CarouselBuilder


CompartmentBuilder
==================

A module for building compartment model systems.

.. autosummary::
   :toctree: generated/

   CompartmentBuilder

"""

from . import carouselBuilder
from .carouselBuilder import *
from .compartmentBuilder import *



import json
from typing import Any

import numpy as np


class NumpyArrayEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for NumPy data types.

    This encoder extends `json.JSONEncoder` to support serialization of NumPy integers,
    floating-point numbers, and arrays. It converts NumPy data types to native Python
    data types that can be serialized to JSON.
    """

    def default(self, obj: Any) -> Any:
        """
        Convert NumPy data types to native Python data types for JSON serialization.

        Parameters
        ----------
        obj : Any
            The object to serialize.

        Returns
        -------
        Any
            The object converted to a JSON-serializable type.

        Raises
        ------
        TypeError
            If the object is not of a supported type.
        """
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyArrayEncoder, self).default(obj)



from __future__ import annotations

import copy
import warnings
from collections import defaultdict
from typing import Any, Callable, Optional, Type

import numpy as np
from addict import Dict
from matplotlib.axes import Axes

from CADETProcess import CADETProcessError, plotting
from CADETProcess.dataStructure import (
    Bool,
    CachedPropertiesMixin,
    Float,
    Integer,
    ParameterBase,
    Sized,
    Structure,
    Typed,
    UnsignedFloat,
    cached_property_if_locked,
    check_nested,
    frozen_attributes,
    generate_nested_dict,
    get_nested_attribute,
    get_nested_value,
)

from .section import MultiTimeLine, Section, TimeLine, generate_indices, unravel

__all__ = ["EventHandler", "Event", "Duration"]


@frozen_attributes
class EventHandler(CachedPropertiesMixin, Structure):
    """
    A handler for dynamic events that affect parameters in a process.

    The `EventHandler` class provides a framework to schedule and manage events
    that cause changes to parameters during a simulation or process. This includes
    single point events as well as durations, and it allows for events to be
    dependent on others, forming complex relationships. Events can be associated
    with transformations or factors that determine their effect.

    Primary functionalities:
    - Schedule events with specific timings and effects.
    - Establish dependencies between events.
    - Manage durations or continuous periods with specific characteristics.
    - Access sorted lists of independent and dependent events.

    Attributes
    ----------
    events : list
        A sorted list of scheduled events, ordered by their execution time.
    durations : list
        List of time durations with specific characteristics.
    event_dict : dict
        A dictionary containing detailed information about all scheduled events.
    durations_dict : dict
        A dictionary containing detailed information about all defined durations.
    independent_events : list
        A list of events that are not influenced by other events.
    dependent_events : list
        A list of events that rely on other events.
    event_performers : dict (not shown in provided code, description based on context)
        A mapping of objects that can perform or be affected by events.
    event_parameters : list
        A list of unique parameters that the events will affect.
    event_times : list
        A list of unique times when events are scheduled to occur, sorted chronologically.
    section_times : list
        A list of times demarcating sections based on event timings.
    n_sections : int
        Total number of sections derived from the section times.
    section_states : dict
        A dictionary providing the state of event parameters at the beginning of every section.
    parameter_events : dict
        A dictionary mapping each parameter to the list of events that affect it.

    Notes
    -----
    The class relies heavily on the concept of "events", which are instances
    of dynamic changes that can influence parameters in the system. These events
    can be independent or based on other events, creating intricate relationships
    to capture complex scenarios.

    See Also
    --------
    Event : Represents a single point change in the system's parameters.
    Duration : Represents a continuous time period with specific attributes or effects.
    """

    cycle_time = UnsignedFloat(default=np.inf)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Construct for EventHandler."""
        super().__init__(*args, **kwargs)

        self._events = []
        self._durations = []
        self._lock = False

    @cached_property_if_locked
    def events(self) -> list[Event]:
        """
        list: All Events ordered by event time.

        See Also
        --------
        Event
        add_event
        remove_event
        event_dependencies
        Durations

        """
        return sorted(self._events, key=lambda evt: evt.time)

    @cached_property_if_locked
    def events_dict(self) -> dict[str, Event | Duration]:
        """Return Events and Durations orderd by name."""
        evts = {evt.name: evt for evt in self.events}
        durs = {dur.name: dur for dur in self.durations}
        return {**evts, **durs}

    def add_event(
        self,
        name: str,
        parameter_path: str,
        state: float,
        time: float = 0.0,
        indices: int = None,
        dependencies: list = None,
        factors: list = None,
        transforms: Optional[list] = None,
    ) -> Event:
        """
        Add a new event that changes a parameter during the process.

        An event is a dynamic alteration that occurs at a specified time and can modify
        the attributes of specific objects involved in the process.

        Parameters
        ----------
        name : str
            Name of the event.
        parameter_path : str
            Path of the parameter that is changed in dot notation.
        state : float
            Value of the attribute that is changed at Event execution.
        time : float
            Time at which the event is executed.
        dependencies : list
            List of the events on which the event time depends.
        factors : List
            List with factors for linear combination of dependencies.
        indices : int
            Index slices for events that modify an entry of a parameter array.
        transforms : list, optional
            List of functions used to transform the parameter value.
            Length must be equal the length of independent events.
            If None, no transform is applied.

        Returns
        -------
        Event:
            The new Event.

        Raises
        ------
        CADETProcessError
            If Event already exists in the event_dict
        CADETProcessError
            If EventPerformer is not found in EventHandler

        See Also
        --------
        events
        remove_event
        add_event_dependency
        Event
        Event.add_dependency
        add_duration
        """
        if name in self.events_dict:
            raise CADETProcessError("Event already exists")
        evt = Event(name, self, parameter_path, state, time=time, indices=indices)

        self._events.append(evt)
        super().__setattr__(name, evt)

        if dependencies is not None:
            self.add_event_dependency(evt.name, dependencies, factors, transforms)

        return evt

    def remove_event(self, evt_name: str) -> None:
        """
        Remove a specified event from the event handler.

        This method ensures that the specified event will no longer influence the
        process by dynamically changing any attributes.

        Parameters
        ----------
        evt_name : str
            Name of the event to be removed

        Raises
        ------
        CADETProcessError
            If Event is not found.

        Notes
        -----
        !!! Check remove_event_dependencies

        See Also
        --------
        add_event
        Event
        Event.remove_dependency
        """
        try:
            evt = self.events_dict[evt_name]
        except KeyError:
            raise CADETProcessError("Event does not exist")

        self._events.remove(evt)
        self.__dict__.pop(evt_name)

    def add_duration(self, name: str, time: float = 0.0) -> "Duration":
        """
        Register a new duration or time point of interest.

        Durations are specific moments in the process that do not necessarily modify
        attributes but are noteworthy or need to be tracked.

        Parameters
        ----------
        name: str
            Name of the event.
        time : float
            Time point for perfoming the event.

        Returns
        -------
        Event:
            The new Event.

        Raises
        ------
        CADETProcessError
            If Duration already exists.

        See Also
        --------
        durations
        remove_duration
        Duration
        add_event
        add_event_dependency
        """
        if name in self.events_dict:
            raise CADETProcessError("Duration already exists")

        dur = Duration(name, self, time)

        self._durations.append(dur)
        super().__setattr__(name, dur)

        return dur

    def remove_duration(self, duration_name: str) -> None:
        """
        Remove a specified duration or time point from tracking.

        This method ensures that the specified duration is no longer considered a point
        of interest in the process.

        Parameters
        ----------
        duration_name : str
            Name of the duration be removed from the EventHandler.

        Raises
        ------
        CADETProcessError
            If Duration is not found.

        See Also
        --------
        add_duration
        Duration
        remove_event_dependency
        """
        try:
            dur = self.events_dict[duration_name]
        except KeyError:
            raise CADETProcessError("Duration does not exist")

        self._durations.remove(dur)
        self.__dict__.pop(duration_name)

    @cached_property_if_locked
    def durations(self) -> list["Duration"]:
        """List of all durations in the process."""
        return self._durations

    def add_event_dependency(
        self,
        dependent_event: str,
        independent_events: list,
        factors: Optional[list] = None,
        transforms: Optional[list] = None,
    ) -> None:
        """
        Create a dependency relationship between events.

        This method establishes how one event (dependent) is influenced by one or more
        other events (independents) through factors and optional transformation
        functions. For example, the time of a dependent event could be determined by the
        sum of the times of independent events multiplied by their corresponding
        factors.

        Parameters
        ----------
        dependent_event : str
            Event whose value will depend on other events.
        independent_events : list
            List of independent event names.
        factors : list, optional
            List of factors used for the relation with the independent events.
            Length must be equal the length of independent events.
            If None, all factors are assumed to be 1.
        transforms : list, optional
            List of functions used to transform the parameter value.
            Length must be equal the length of independent events.
            If None, no transform is applied.

        Raises
        ------
        CADETProcessError
            If dependent_event OR independent_events are not found.
            If length of factors does not equal length of independent events.
            If length of transforms does not equal length of independent events.

        See Also
        --------
        Event
        add_event
        add_duration
        remove_event_dependency
        """
        try:
            evt = self.events_dict[dependent_event]
        except KeyError:
            raise CADETProcessError("Cannot find dependent Event")

        if not isinstance(independent_events, list):
            independent_events = [independent_events]
        if not all(indep in self.events_dict for indep in independent_events):
            raise CADETProcessError("Cannot find one or more independent events")

        if factors is None:
            factors = [1] * len(independent_events)

        if not isinstance(factors, list):
            factors = [factors]
        if len(factors) != len(independent_events):
            raise CADETProcessError(
                "Length of factors must equal length of independent Events"
            )
        if transforms is None:
            transforms = [None] * len(independent_events)

        if not isinstance(transforms, list):
            transforms = [transforms]
        if len(transforms) != len(independent_events):
            raise CADETProcessError(
                "Length of transforms must equal length of independent Events"
            )

        for indep, fac, trans in zip(independent_events, factors, transforms):
            indep = self.events_dict[indep]
            evt.add_dependency(indep, fac, trans)

    def remove_event_dependency(
        self, dependent_event: str, independent_events: list
    ) -> None:
        """
        Remove a previously defined dependency between events.

        Parameters
        ----------
        dependent_event : str
            Name of the event whose value will depend on other events.
        independent_events : list
            List of independent event names.

        Raises
        ------
        CADETProcessError
            If dependent_event is not in list events.
            If one or more independent event is not in list events and
            durations.

        See Also
        --------
        Event
        Event.remove_dependecy
        add_event_dependency

        """
        if dependent_event not in self.events:
            raise CADETProcessError("Cannot find dependent Event")

        if not all(evt in self.events_dict for evt in independent_events):
            raise CADETProcessError("Cannot find one or more independent events")

        for indep in independent_events:
            self.events[dependent_event].remove_dependency(indep)

    @cached_property_if_locked
    def independent_events(self) -> list[Event]:
        """list: All events that are not dependent on other events."""
        return list(filter(lambda evt: evt.is_independent, self.events))

    @cached_property_if_locked
    def dependent_events(self) -> list[Event]:
        """list: All events that are dependent on other events."""
        return list(filter(lambda evt: evt.is_independent is False, self.events))

    @cached_property_if_locked
    def event_parameters(self) -> list[str]:
        """list: Event parameters."""
        return list({evt.parameter_path for evt in self.events})

    @cached_property_if_locked
    def event_performers(self) -> list[str]:
        """list: Event peformers."""
        return list({evt.performer for evt in self.events})

    @cached_property_if_locked
    def event_times(self) -> list[float]:
        """list: Time of events, sorted by Event time."""
        event_times = list({evt.time for evt in self.events})
        event_times.sort()

        return event_times

    @cached_property_if_locked
    def section_times(self) -> list[float]:
        """
        list: Section times.

        Includes 0 and cycle_time if they do not coincide with event time.

        """
        if len(self.event_times) == 0:
            return [0, self.cycle_time]

        section_times = self.event_times

        if section_times[0] != 0:
            section_times = [0] + section_times
        if section_times[-1] != self.cycle_time:
            section_times = section_times + [self.cycle_time]

        return section_times

    @property
    def n_sections(self) -> int:
        """int: Number of sections."""
        return len(self.section_times) - 1

    @cached_property_if_locked
    def section_states(self) -> Dict[float, dict[float, np.ndarray]]:
        """Return state of event parameters at every section."""
        parameter_timelines = self.parameter_timelines
        section_states = defaultdict(dict)

        for sec_time in self.section_times[0:-1]:
            for param, tl in parameter_timelines.items():
                section_states[sec_time][param] = tl.coefficients(sec_time)

        return Dict(section_states)

    @cached_property_if_locked
    def parameter_events(self) -> dict[str, list[Event] | dict[int, list[Event]]]:
        """
        Return event parameters mapped to their corresponding events.

        This dictionary associates each event parameter with its list of events.
        For events that are index-specific, an inner dictionary is used, where
        each index maps to its list of events.

        Notes
        -----
        For index-dependent events, a separate key is added for each index.
        """
        parameter_events = defaultdict(list)
        for evt in self.events:
            if evt.is_index_specific:
                for index in evt.full_indices:
                    parameter_events[evt.parameter_path] = defaultdict(list)

        for evt in self.events:
            if evt.is_index_specific:
                for index in evt.full_indices:
                    parameter_events[evt.parameter_path][index].append(evt)
            else:
                parameter_events[evt.parameter_path].append(evt)
        return Dict(parameter_events)

    @cached_property_if_locked
    def parameter_timelines(self) -> Dict[str, TimeLine]:
        """
        Return Dict: TimeLine representation for every event parameter.

        This dictionary associates each event parameter with its TimeLine object.
        If an event parameter is considered as one of the 'sized parameters',
        it gets associated with a MultiTimeLine object, which handles
        multi-dimensional or indexed data.

        Each timeline, be it a regular or multi-timeline, consists of
        sections representing time intervals where the parameter holds
        a specific value or state.
        """
        parameter_timelines = {}
        multi_timelines = {}

        parameters = self.parameters
        for param in self.event_parameters:
            if param not in self.sized_parameters:
                parameter_timelines[param] = TimeLine()
            else:
                base_state = get_nested_value(parameters, param)
                is_polynomial = check_nested(self.polynomial_parameters, param)
                multi_timelines[param] = MultiTimeLine(base_state, is_polynomial)

        for evt_parameter, events in self.parameter_events.items():
            if not isinstance(events, dict):
                events = {None: events}

            for index, index_events in events.items():
                for i_evt, evt in enumerate(index_events):
                    section_start = evt.time

                    if i_evt < len(index_events) - 1:
                        section_end = index_events[i_evt + 1].time
                        self._create_and_add_sections(
                            section_start,
                            section_end,
                            evt,
                            index,
                            parameter_timelines,
                            multi_timelines,
                        )
                    else:
                        section_end = self.cycle_time
                        self._create_and_add_sections(
                            section_start,
                            section_end,
                            evt,
                            index,
                            parameter_timelines,
                            multi_timelines,
                        )

                        if index_events[0].time != 0:
                            section_start = 0.0
                            section_end = index_events[0].time
                            self._create_and_add_sections(
                                section_start,
                                section_end,
                                evt,
                                index,
                                parameter_timelines,
                                multi_timelines,
                            )

        for param, tl in multi_timelines.items():
            parameter_timelines[param] = tl.combined_time_line

        return Dict(parameter_timelines)

    def _create_and_add_sections(
        self,
        start: float,
        end: float,
        evt: Event,
        index: int | tuple[int],
        parameter_timelines: dict,
        multi_timelines: dict,
    ) -> None:
        """
        Create a new Section and integrate it into the correct TimeLine.

        This method forms a new `Section` object using the provided start and end times,
        and the state from the event `evt`. Depending on whether the event is index-
        specific, this section is then added to a regular TimeLine or a MultiTimeLine.

        Parameters
        ----------
        start : float
            Starting time of the Section.
        end : float
            Ending time of the Section.
        evt : Event
            Event associated with the Section.
            Determines the state for this time interval.
        index : int or tuple
            Index or indices pointing to specific entries in indexed event parameters.
        parameter_timelines : dict
            Dictionary mapping parameter names to their respective TimeLine objects.
        multi_timelines : dict
            Dictionary mapping parameter names to their respective MultiTimeLine objects.
        """
        if not evt.is_index_specific:
            section = Section(start, end, evt.full_state)
            parameter_timelines[evt.parameter_path].add_section(section)
        else:
            section = Section(
                start, end, evt.index_states[index], is_polynomial=evt.is_polynomial
            )
            if evt.degree > 0 and len(index) == 1:
                index = (0,) + index
            multi_timelines[evt.parameter_path].add_section(section, index)

    @property
    def performer_events(self) -> Dict[str, list[Event]]:
        """
        Return Dict: Event performer mapped to their corresponding list of events.

        For every event, this dictionary associates the event's performer
        with the event. This allows for easy retrieval of all events carried out
        by a specific performer.
        """
        performer_events = defaultdict(list)
        for evt in self.events:
            performer_events[evt.performer].append(evt)

        return Dict(performer_events)

    @cached_property_if_locked
    def performer_timelines(self) -> Dict[str, dict[str, TimeLine]]:
        """
        Return Dict: Each performer mapped to their TimeLines based on event parameters.

        This dictionary provides a representation of event parameters in the form
        of timelines for each performer. This hierarchical structure helps in
        quickly accessing the TimeLine of any event parameter specific to a performer.
        """
        performer_timelines = {performer: {} for performer in self.event_performers}

        for param, tl in self.parameter_timelines.items():
            performer, param = param.rsplit(".", 1)
            performer_timelines[performer][param] = tl

        return performer_timelines

    @property
    def parameters(self) -> dict:
        """
        dict: The EventHandler parameters.

        In addition to the standard parameters retrieved from the superclass,
        this property adds event parameters from independent events, parameters
        from durations, and the cycle time.
        """
        parameters = super().parameters

        events = {evt.name: evt.parameters for evt in self.independent_events}
        parameters.update(events)

        events = {evt.name: evt.parameters for evt in self.dependent_events}
        parameters.update(events)

        durations = {dur.name: dur.parameters for dur in self.durations}
        parameters.update(durations)

        parameters["cycle_time"] = self.cycle_time

        return parameters

    @parameters.setter
    def parameters(self, parameters: dict) -> None:
        """Set event parameters based on provided dictionary."""
        try:
            self.cycle_time = parameters.pop("cycle_time")
        except KeyError:
            pass

        for evt_name, evt_parameters in parameters.items():
            try:
                evt = self.events_dict[evt_name]
            except AttributeError:
                raise CADETProcessError("Not a valid event")
            if (
                "time" in evt_parameters
                and evt not in self.independent_events + self.durations
            ):
                raise CADETProcessError(
                    f'Cannot set "time" for {str(evt)} because it is not an independent event.'
                )

            evt.parameters = evt_parameters

    @property
    def sized_parameters(self) -> dict:
        """
        dict: Compilation of parameters from events with indices.

        Besides the sized parameters fetched from the superclass, this property
        collects parameters from events that have associated indices.
        """
        parameters = super().sized_parameters

        events = {
            evt.parameter_path: evt.parameters
            for evt in self.events
            if evt.indices is not None
        }
        parameters.update(events)

        return parameters

    def check_config(self) -> bool:
        """
        Validate the event configuration.

        Ensure no duplicate events exist for a specific parameter and index and
        verify that constants are incorporated in polynomials.

        Returns
        -------
        bool
            True if all validations pass, False otherwise.
        """
        flag = True

        if not self.check_duplicate_events():
            flag = False

        if not self.check_uninitialized_indices():
            flag = False

        return flag

    def check_duplicate_events(self) -> bool:
        """
        Ensure no simulateneous events are scheduled for a specific parameter and index.

        Evaluates all events scheduled for each parameter and index combination.
        Raises a warning if multiple events are scheduled to occur simultaneously,
        as this can lead to unexpected system or simulation behavior.

        Returns
        -------
        bool
            True if no duplicate events are detected, False otherwise.

        Warnings
        --------
            If events are detected to occur at the same timestamp.
        """
        flag = True
        for evt_parameter, events in self.parameter_events.items():
            if not isinstance(events, dict):
                events = {None: events}

            for index, index_events in events.items():
                index_event_times = [evt.time for evt in index_events]

                duplicates = [
                    time
                    for time in set(index_event_times)
                    if index_event_times.count(time) > 1
                ]

                if duplicates:
                    duplicate_events = [
                        evt for evt in index_events if evt.time in duplicates
                    ]
                    warnings.warn(
                        f"Got multiple events at the same time: {duplicate_events}"
                    )
                    flag = False

        return flag

    def check_uninitialized_indices(self) -> bool:
        """
        Ensure all indices are specified when a parameter isn't initialized.

        Returns
        -------
        bool
            True if all indices are properly defined, False otherwise.

        Warnings
        --------
            If there are parameters with uninitialized entries for some indices.
        """
        flag = True
        for evt_parameter, events in self.parameter_events.items():
            current_value = get_nested_value(self.parameters, evt_parameter)
            current_value = np.array(current_value)

            if np.any(np.isnan(current_value)):
                warnings.warn(f"{evt_parameter} has entries which were not initialized")
                flag = False

        return flag

    def plot_events(self, x_axis_in_minutes: bool = True) -> list[Axes]:
        """
        Plot parameter state as a function of time.

        The method creates a plot for each parameter timeline and displays the state
        of the parameter against time. The time is represented on the x-axis, while
        the parameter state is shown on the y-axis.

        Parameters
        ----------
        x_axis_in_minutes: bool, optional
            If True, the x-axis will be plotted using minutes. The default is True.

        Returns
        -------
        list[Axes]
            List of Axes objects, each containing a plot of the parameter state.

        Notes
        -----
        The time is divided into 1001 linearly spaced points between 0 and the cycle
        time for the evaluation of the parameter state.
        """
        time_s = np.linspace(0, self.cycle_time, 1001)

        time_ax = time_s
        if x_axis_in_minutes:
            time_ax = time_ax / 60

        axs: list[Axes] = []

        for parameter, tl in self.parameter_timelines.items():
            fig, ax = plotting.setup_figure()

            y = tl.value(time_s)

            layout = plotting.Layout()
            layout.title = str(parameter)
            layout.x_label = "$time~/~s$"
            if x_axis_in_minutes:
                layout.x_label = "$time~/~min$"
            layout.y_label = "$state$"

            ax.plot(time_ax, y)

            plotting.set_layout(ax, layout)

            axs.append(ax)

        return axs


class Event:
    """
    Defines dynamic changes of model parameters based on events.

    An `Event` is a time-based modification to an attribute of a performer.
    Its execution time can depend on other Events or Durations. To handle
    cyclic behavior, times are computed modulo the cycle time of the EventHandler.

    Attributes
    ----------
    name : str
        The event's name.
    event_handler : EventHandler
        Object managing the performers and cycle time.
    parameter_path : str
        Dot notation path to the target parameter within the evaluation_object.
    state : float
        Desired attribute value when the event is executed.
    time : float, default=0.0
        The execution time of the event.
    indices : int or list, default=None
        Specific indices if the event modifies a parameter array entry.

    See Also
    --------
    EventHandler
    Duration
    """

    _parameters = ["time", "state"]

    def __init__(
        self,
        name: str,
        event_handler: EventHandler,
        parameter_path: str,
        state: float,
        time: float = 0.0,
        indices: Optional[int | list[int]] = None,
    ) -> None:
        """
        Initialize the Event object.

        Parameters
        ----------
        name : str
            The event's name.
        event_handler : EventHandler
            Object managing the performers and cycle time.
        parameter_path : str
            Dot notation path to the target parameter within the evaluation_object.
        state : float
            Desired attribute value when the event is executed.
        time : float, default=0.0
            The execution time of the event.
        indices : int or list of int, optional
            Specific indices if the event modifies a parameter array entry.
            Can also accept slices.
        """
        self.name = name

        self.event_handler = event_handler
        self.parameter_path = parameter_path

        self.indices = indices
        self.state = state

        self._dependencies = []
        self._factors = []
        self._transforms = []

        self.time = time

    @property
    def parameter_path(self) -> str:
        """str: Dot notation path to the target parameter within the evaluation_object."""
        return self._parameter_path

    @parameter_path.setter
    def parameter_path(self, parameter_path: str) -> None:
        if not check_nested(
            self.event_handler.section_dependent_parameters, parameter_path
        ):
            raise CADETProcessError("Not a valid event parameter")
        self._parameter_path = parameter_path

    @property
    def parameter_sequence(self) -> tuple[str, ...]:
        """tuple: Tuple of parameters path elements."""
        return tuple(self.parameter_path.split("."))

    @property
    def parameter_descriptor(self) -> Optional[ParameterBase]:
        """Return parameter descriptor."""
        performer_class = type(self.performer_obj)
        try:
            descriptor = getattr(performer_class, self.parameter_sequence[-1])
        except AttributeError:
            return None

        if not isinstance(descriptor, ParameterBase):
            return None

        return descriptor

    @property
    def parameter_type(self) -> Type[Any]:
        """type: Type of the parameter."""
        if isinstance(self.parameter_descriptor, Typed):
            return self.parameter_descriptor.ty

        if self.current_value is None:
            raise CADETProcessError(
                "Parameter is not initialized. Cannot determine parameter type."
            )

        return type(self.current_value)

    @property
    def parameter_shape(self) -> tuple[int, ...]:
        """tuple: Shape of the parameter array."""
        param_descriptor = self.parameter_descriptor
        if isinstance(param_descriptor, (Float, Integer, Bool)):
            return ()

        if isinstance(param_descriptor, Sized):
            shape = param_descriptor.get_expected_size(self.performer_obj)
            if not isinstance(shape, tuple):
                shape = (shape,)

            return shape

        cur_value = self.current_value
        if cur_value is None:
            raise CADETProcessError(
                "Parameter is not initialized. Cannot determine parameter shape."
            )

        return np.array(cur_value).shape

    @property
    def is_sized(self) -> bool:
        """bool: True if descriptor is instance of Sized. False otherwise."""
        if isinstance(self.parameter_descriptor, (Float, Integer, Bool)):
            return False

        if isinstance(self.parameter_descriptor, Sized):
            return True

        if self.current_value is None:
            raise CADETProcessError(
                "Parameter is not initialized. "
                "Cannot determine dimensions required for setting index."
            )
        return np.array(self.current_value).size > 1

    @property
    def is_polynomial(self) -> bool:
        """bool: True if descriptor is instance of NdPolynomial. False otherwise."""
        return check_nested(
            self.event_handler.polynomial_parameters, self.parameter_path
        )

    @property
    def degree(self) -> int:
        """int: The degree of the polynomial event state."""
        if self.is_polynomial:
            shape = self.parameter_shape
            return shape[-1] - 1
        else:
            return 0

    @property
    def indices(self) -> list[tuple[int]] | None:
        """
        list: Indices for events that modifies only specific entries of a parameter.

        List of tuples for each entry. If parameter is scalar, None
        """
        param_shape = self.parameter_shape

        if len(param_shape) == 0:
            return

        indices = generate_indices(param_shape, self._indices)

        # Check if all indices unique:
        full_indices = unravel(param_shape, indices)
        duplicates = [
            index for index in set(full_indices) if full_indices.count(index) > 1
        ]

        if len(duplicates) > 0:
            raise ValueError(f"Got duplicate entries for indices {duplicates}")

        return indices

    @indices.setter
    def indices(self, indices: list[int]) -> None:
        """
        list: Indices of parameters to set Event state.

        Can be list of tuples. Including slicing.
        """
        if indices is not None and not self.is_sized:
            raise IndexError("Events for scalar parameters cannot have indices.")

        self._indices = indices

        # Since indices are constructed on `get`, call the property here:
        try:
            _ = self.indices
        except (ValueError, TypeError) as e:
            raise e

    @property
    def is_index_specific(self) -> bool:
        """bool: True if event modifies entry of a parameter array, False otherwise."""
        if len(self.full_indices) > 0:
            return True
        else:
            return False

    @property
    def full_indices(self) -> list[int]:
        """list: Full indices."""
        indices = self.indices
        if self.indices is None and len(self.parameter_shape) > 0:
            indices = generate_indices(self.parameter_shape)
        return unravel(self.parameter_shape, indices)

    @property
    def n_indices(self) -> int:
        """int: Number of (full) indices."""
        if len(self.parameter_shape) > 0:
            return len(self.full_indices)
        else:
            return 0

    @property
    def n_entries(self) -> int:
        """int: The number of entries in the event state."""
        if self.is_polynomial:
            return np.array(self.full_state).shape[0]
        else:
            if isinstance(self.full_state, (int, float, bool)):
                return 1
            else:
                return self.n_indices

    def add_dependency(
        self,
        dependency: Event,
        factor: float = 1,
        transform: Optional[Callable] = None,
    ) -> None:
        """
        Add a time dependency on another event.

        When an event is dependent, the time of the event is based on a linear
        combination of its dependencies.

        Parameters
        ----------
        dependency : Event
            Event that this event depends on.
        factor : float, default=1
            Weighting factor for the dependency.
        transform : callable, optional
            A function to transform the dependent event's time.

        Raises
        ------
        CADETProcessError
            If the dependency is already listed.
        """
        if dependency in self._dependencies:
            raise CADETProcessError("Dependency already exists")

        self._dependencies.append(dependency)
        self._factors.append(factor)
        if transform is None:

            def transform(t: Any) -> Any:
                return t

        self._transforms.append(transform)

    def remove_dependency(self, dependency: Event) -> None:
        """
        Remove dependencies of events.

        Parameters
        ----------
        dependency : Event
            Event object to remove from dependencies.

        Raises
        ------
        CADETProcessError
            If the dependency doesn't exists in list dependencies.
        """
        if dependency in self._dependencies:
            raise CADETProcessError("Dependency not found")

        index = self._dependencies(dependency)

        del self._dependencies[index]
        del self._factors[index]
        del self._transforms[index]

    @property
    def dependencies(self) -> list[Event]:
        """list: Events on which the Event depends."""
        return self._dependencies

    @property
    def is_independent(self) -> bool:
        """bool: True, if event is independent, False otherwise."""
        if len(self.dependencies) == 0:
            return True
        else:
            return False

    @property
    def factors(self) -> list[int]:
        """list: Linear coefficients for dependent events."""
        return self._factors

    @property
    def transforms(self) -> list[Callable]:
        """list: Transform functions for dependent events."""
        return self._transforms

    @property
    def time(self) -> float:
        """
        float: Time when the event is executed.

        If the value is larger than the cycle time, the time modulo cycle time
        is returned. If the Event is not independent, the time is calculated
        from its dependencies.

        Raises
        ------
        CADETProcessError
            If the event is not independent.

        """
        if self.is_independent:
            time = self._time
        else:
            transformed_time = [
                f(dep.time) for f, dep in zip(self.transforms, self.dependencies)
            ]
            time = np.dot(transformed_time, self._factors)
        cycle_time = getattr(self.event_handler, "cycle_time")
        return time % cycle_time

    @time.setter
    def time(self, time: float) -> None:
        if not np.isscalar(time):
            raise TypeError("Expected scalar value")

        if self.is_independent:
            self._time = time
        else:
            raise CADETProcessError("Cannot set time for dependent events")

    @property
    def state(self) -> float | np.ndarray:
        """
        Return the state of the parameter event.

        When retrieving, it returns the current state of the event.
        When setting, the internal state is updated.

        Returns
        -------
        float | np.ndarray
            The state of the parameter event.
        """
        return self._state

    @state.setter
    def state(self, state: float | np.ndarray) -> None:
        """
        Set the state of the event.

        If indices are not defined and there's no current value, it initializes
        the state with the provided value. The state is then updated with the
        calculated full state based on the indices and provided value.

        Parameters
        ----------
        state : float or np.ndarray
            Value to set as the new state of the event.

        Raises
        ------
        ValueError, TypeError
            If the updated state does not align with the expected data type or structure.
        """
        # Initialize value to get dimensions
        if self._indices is None and self.current_value is None:
            self.set_value(state)
            state = self.current_value

        self._state = state

        # Since event is constructed on `get`, call the property here:
        try:
            _ = self.full_state
        except (ValueError, TypeError) as e:
            raise e

    def _ensure_2D_for_slices(self, state: float | np.ndarray) -> None:
        """
        Ensure the state is 2D when dealing with slices.

        If there's only one set of indices and it contains a slice, it
        prepares the state to be handled as a 2D structure.

        Parameters
        ----------
        state : float or np.ndarray
            The state that might need to be converted.

        Returns
        -------
        float or np.ndarray
            Original state or a 2D structure depending on the indices.
        """
        if len(self.indices) == 1 and any(
            isinstance(i, slice) for i in self.indices[0]
        ):
            state = [state]

        return state

    @property
    def full_state(self) -> float | list:
        """
        Construct the full state based on indices and current value.

        This method reconstructs the state from the stored state value,.

        Returns
        -------
        float or list
            The computed full state, either as a scalar or an array.

        Raises
        ------
        ValueError
            If the length of the state does not match the length of the indices.
        """
        state = self._state
        indices = self.indices

        # Get new (full) parameter value
        if self._indices is None:
            new_value = state
        else:
            if self.current_value is None:
                new_value = np.full(self.parameter_shape, np.nan)
            else:
                new_value = np.array(self.current_value, ndmin=1)

            # Ensure state is list
            if not isinstance(state, list):
                state = [state]

            # Ensure state is 2D for indices that contain slices
            state = self._ensure_2D_for_slices(state)

            if len(state) != len(indices):
                raise ValueError(
                    f"Expected {len(self.indices)} entries for state. Got {len(state)}"
                )

            for i, ind in enumerate(indices):
                expected_shape = new_value[ind].shape
                if (
                    self.is_polynomial
                    and len(self.parameter_shape) > 1
                    and len(ind) == 1
                ):
                    new_slice = self.parameter_descriptor.fill_values(
                        expected_shape, state[i]
                    )
                else:
                    new_slice = np.array(state[i], ndmin=1)

                if any(isinstance(i, slice) for i in ind):
                    if new_slice.size != np.prod(expected_shape):
                        new_slice = np.broadcast_to(new_slice, expected_shape)
                    else:
                        new_slice = np.reshape(new_slice, expected_shape)

                if len(expected_shape) == 0:
                    new_slice = new_slice.squeeze()
                new_value[ind] = new_slice

            if self.parameter_type is not np.ndarray:
                new_value = self.parameter_type(new_value.tolist())

        # Set the value:
        self.set_value(new_value)
        new_value = self.current_value

        if indices is not None:
            new_value = np.array(new_value, ndmin=1)
            full_state = []
            for ind in indices:
                full_state += new_value[ind].flatten().tolist()
        else:
            full_state = new_value

        return full_state

    @property
    def index_states(self) -> dict[tuple, float]:
        """dict[tuple, float]: State values mapped to their respective indices."""
        index_states = {}
        for ind, state in zip(self.full_indices, self.full_state):
            index_states[ind] = state

        return index_states

    @property
    def performer(self) -> str:
        """str: The name of the performer of the event."""
        if len(self.parameter_sequence) == 1:
            return self.parameter_sequence[0]
        else:
            return ".".join(self.parameter_sequence[:-1])

    @property
    def performer_obj(self) -> Any:
        """any: Performer object from the event handler."""
        return get_nested_attribute(self.event_handler, self.performer)

    def set_value(self, state: float | np.ndarray) -> None:
        """Set the specified state to the associated event parameter."""
        state = copy.deepcopy(state)
        if self.parameter_descriptor is not None:
            setattr(self.performer_obj, self.parameter_sequence[-1], state)
        else:
            parameters = generate_nested_dict(self.parameter_sequence, state)
            self.event_handler.parameters = parameters

    @property
    def current_value(self) -> Any:
        """any: Current state of the associated event parameter."""
        if self.parameter_descriptor is not None:
            return getattr(self.performer_obj, self.parameter_sequence[-1])
        else:
            return get_nested_value(self.event_handler.parameters, self.parameter_path)

    @property
    def parameters(self) -> dict:
        """dict: list with all parameters."""
        return Dict({param: getattr(self, param) for param in self._parameters})

    @parameters.setter
    def parameters(self, parameters: float | int | dict) -> None:
        if isinstance(parameters, (float, int)):
            self.time = parameters
        else:
            for param, value in parameters.items():
                if param not in self._parameters:
                    raise CADETProcessError("Not a valid parameter")
                setattr(self, param, value)

    def __repr__(self) -> str:
        """str: String representation of the Event."""
        representation = (
            f"{self.__class__.__name__}("
            f"name={self.name}, "
            f"parameter_path={self.parameter_path}, "
            f"state={self.state}, "
            f"time={self.time}"
        )
        if self.indices is not None:
            representation += f", indices={self.indices}"

        representation += ")"

        return representation


class Duration:
    """
    Class for representing a duration between two events in an Eventhandler.

    Attributes
    ----------
    start_event : str
        Name of the start event of a duration.
    end_event : str
        Name of the end event of a duration.
    """

    def __init__(
        self,
        name: str,
        event_handler: EventHandler,
        time: float = 0.0,
    ) -> None:
        """Initialize Duration Object."""
        self.name = name
        self.time = time
        self._parameters = ["time"]

    @property
    def parameters(self) -> Dict:
        """dict: list with all parameters."""
        return Dict({param: getattr(self, param) for param in self._parameters})

    @parameters.setter
    def parameters(self, parameters: float | int | dict) -> None:
        if isinstance(parameters, (float, int)):
            self.time = parameters
        else:
            for param, value in parameters.items():
                if param not in self._parameters:
                    raise CADETProcessError("Not a valid parameter")
                setattr(self, param, value)

    def __repr__(self) -> str:
        """str: String representation of the Duration."""
        return f"{self.__class__.__name__}(name={self.name}, time={self.time})"



import unittest

from CADETProcess.processModel import (
    ComponentSystem,
    Cstr,
    FlowSheet,
    GeneralRateModel,
    Inlet,
    LumpedRateModelWithoutPores,
    LumpedRateModelWithPores,
    MassActionLaw,
    MassActionLawParticle,
    Outlet,
    Process,
    TubularReactor,
)
from CADETProcess.simulator import Cadet

from tests.test_cadet_adapter import found_cadet


def setup_process(unit_type):
    component_system = ComponentSystem(2)

    inlet = Inlet(component_system, "inlet")
    inlet.c = [1, 0]
    inlet.flow_rate = 1e-3

    if unit_type == "cstr":
        cstr = Cstr(component_system, "reaction_unit")
        total_volume = 1e-3
        total_porosity = 0.7
        cstr.init_liquid_volume = total_porosity * total_volume
        cstr.const_solid_volume = (1 - total_porosity) * total_volume

        unit = cstr
    elif unit_type == "pfr":
        pfr = TubularReactor(component_system, "reaction_unit")
        pfr.length = 1
        pfr.diameter = 1e-2
        pfr.axial_dispersion = 1e-6

        unit = pfr
    elif unit_type == "lrm":
        lrm = LumpedRateModelWithoutPores(component_system, "reaction_unit")
        lrm.length = 1
        lrm.diameter = 1e-2
        lrm.total_porosity = 0.7
        lrm.axial_dispersion = 1e-6

        unit = lrm
    elif unit_type == "lrmp":
        lrmp = LumpedRateModelWithPores(component_system, "reaction_unit")
        lrmp.length = 1
        lrmp.diameter = 1e-2
        lrmp.bed_porosity = 0.7
        lrmp.particle_radius = 1e-6
        lrmp.particle_porosity = 0.7
        lrmp.axial_dispersion = 1e-6
        lrmp.film_diffusion = 2 * [1e-3]

        unit = lrmp
    elif unit_type == "grm":
        grm = GeneralRateModel(component_system, "reaction_unit")

        grm.length = 1
        grm.diameter = 1e-2
        grm.bed_porosity = 0.7
        grm.particle_radius = 1e-6
        grm.particle_porosity = 0.7
        grm.axial_dispersion = 1e-6
        grm.film_diffusion = 2 * [1e-3]
        grm.pore_diffusion = 2 * [1e-3]

        unit = grm
    else:
        raise ValueError("Unknown Model.")

    outlet = Outlet(component_system, "outlet")

    flow_sheet = FlowSheet(component_system)

    flow_sheet.add_unit(inlet)
    flow_sheet.add_unit(unit)
    flow_sheet.add_unit(outlet)

    flow_sheet.add_connection(inlet, unit)
    flow_sheet.add_connection(unit, outlet)

    process = Process(flow_sheet, "test_reaction_bulk")
    process.cycle_time = 10

    return process


class TestReaction(unittest.TestCase):
    """Unit tests for the reaction module."""

    @unittest.skipIf(found_cadet is False, "Skip if CADET is not installed.")
    def test_reaction_bulk(self):
        """Test the reaction in bulk."""
        for unit_type in ["cstr", "pfr", "lrmp", "grm"]:
            with self.subTest(unit_type=unit_type):
                process = setup_process(unit_type)

                bulk_reaction_model = MassActionLaw(process.component_system)
                bulk_reaction_model.add_reaction([0, 1], [-1, 1], 1)

                process.flow_sheet.reaction_unit.bulk_reaction_model = (
                    bulk_reaction_model
                )

                simulator = Cadet()
                results = simulator.simulate(process)

                self.assertTrue(results.solution.outlet.inlet.solution[-1, :][0] != 1)
                self.assertTrue(results.solution.outlet.inlet.solution[-1, :][1] != 0)

    @unittest.skipIf(found_cadet is False, "Skip if CADET is not installed.")
    def test_reaction_particle(self):
        """Test the reaction in particle.

        Notes
        -----
        Cstr currently not working. Need to investigate

        """
        for unit_type in ["lrm", "lrmp", "grm"]:
            with self.subTest(unit_type=unit_type):
                process = setup_process(unit_type)

                particle_reaction_model = MassActionLawParticle(
                    process.component_system
                )
                particle_reaction_model.add_liquid_reaction([0, 1], [-1, 1], 1)

                process.flow_sheet.reaction_unit.particle_reaction_model = (
                    particle_reaction_model
                )

                simulator = Cadet()
                results = simulator.simulate(process)

                self.assertTrue(results.solution.outlet.inlet.solution[-1, :][0] != 1)
                self.assertTrue(results.solution.outlet.inlet.solution[-1, :][1] != 0)

    @unittest.skipIf(found_cadet is False, "Skip if CADET is not installed.")
    def test_reaction_bulk_in_particle(self):
        """Test the reaction in particle when setting a bulk reaction type.

        Notes
        -----
        Cstr currently not working. Need to investigate

        """
        for unit_type in ["lrm", "lrmp", "grm"]:
            with self.subTest(unit_type=unit_type):
                process = setup_process(unit_type)

                bulk_reaction_model = MassActionLaw(process.component_system)
                bulk_reaction_model.add_reaction([0, 1], [-1, 1], 1)

                process.flow_sheet.reaction_unit.particle_reaction_model = (
                    bulk_reaction_model
                )

                simulator = Cadet()
                results = simulator.simulate(process)

                self.assertTrue(results.solution.outlet.inlet.solution[-1, :][0] != 1)
                self.assertTrue(results.solution.outlet.inlet.solution[-1, :][1] != 0)


if __name__ == "__main__":
    unittest.main()



import copy
import math
import operator
import typing as tp
from typing import Any, Optional, Union

import numpy as np
import numpy.typing as npt

from .dataStructure import Descriptor


class ParameterBase(Descriptor):
    """
    Base class for model parameters with potential constraints or type-casting.

    Unlike mere data members, parameters can have default values, support type
    constraints, and cast from certain types to their target type.

    Attributes
    ----------
    default : Any
        Default value of the parameter.
    unit : str
        Unit of the parameter.
    description : str
        Description or context of the parameter.

    Notes
    -----
    1. Supports deep copying of default values, allowing mutable defaults without side effects.
    2. Subclasses can further specify type constraints (like `Typed`).
    3. They can also define
      - immutable parameters (like `Constant`) and
      - options-based parameters (`Switch`).

    See Also
    --------
    Descriptor
    Structure
    Constant
    Switch
    Typed
    Bool
    Integer
    Tuple
    Float
    String
    Dictionary
    """

    def __init__(
        self,
        *args: Any,
        default: Optional[Any] = None,
        is_optional: bool = False,
        unit: Optional[str] = None,
        description: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Parameter instance.

        Parameters
        ----------
        *args : Any
            Variable length argument list.
        default : Any, optional
            Default value for the parameter. Defaults to None.
        is_optional : bool, optional
            If True, parameter is not added to list of required parameters.
            Defaults to False
        unit : str, optional
            Unit of the parameter. Defaults to None.
        description : str, optional
            Description of the parameter. Defaults to None.
        **kwargs : Any
            Arbitrary keyword arguments.
        """
        self.default = default
        self.is_optional = is_optional
        self.unit = unit
        self.description = description
        super().__init__(*args, **kwargs)

    @property
    def default(self) -> Any:
        """Any: Get or set the default value of the parameter."""
        return copy.deepcopy(self._default)

    @default.setter
    def default(self, value: Any) -> None:
        """
        Set the default value of the parameter.

        Parameters
        ----------
        value : Any
            Value to set as default.
        """
        if value is not None:
            val = self._prepare(None, value, recursive=True)
            self._check(None, val, recursive=True)

        self._default = value

    def get_default_value(self, instance: Any) -> Any:
        """
        Return default values if necessary.

        Override this method if type-casting for default values is necessary.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the default value for.

        Returns
        -------
        Any
            Default value.
        """
        default = self.default

        if default is not None:
            default = self._prepare(instance, default, recursive=True)
            if default is not None:
                self._check(instance, default, recursive=True)

        return default

    def __get__(self, instance: Any, cls: type) -> Any:
        """
        Retrieve the descriptor value for the given instance.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        cls : Type[Any]
            Class to which the descriptor belongs.

        Returns
        -------
        Any
            Descriptor value or the default value if the descriptor value isn't set.
        """
        if instance is None:
            return self

        try:
            value = Descriptor.__get__(self, instance, cls)
        except KeyError:
            value = self.get_default_value(instance)

        if value is not None:
            self._check(instance, value, recursive=True)

        return value

    def __set__(self, instance: Any, value: Any) -> None:
        """
        Set the descriptor value for the given instance.

        Parameters
        ----------
        instance : Any
            Instance to set the descriptor value for.
        value : Any
            Value to set.
        """
        if value is None:
            value = self.get_default_value(instance)

        if value is not None:
            value = self._prepare(instance, value, recursive=True)
            self._check(instance, value, recursive=True)

        try:
            if self.name in instance._parameters:
                instance._parameters_dict[self.name] = value
        except AttributeError:
            pass

        super().__set__(instance, value)

    def _prepare(self, instance: Any, value: Any, recursive: bool = False) -> Any:
        """
        Prepare value for setting if necessary.

        Override this method if type-casting or other operations are necessary.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Any
            Value to cast.
        recursive : bool, optional
            If True, prepare values recursively.

        Returns
        -------
        Any
            Prepared value.
        """
        return value

    def _check(self, instance: Any, value: Any, recursive: bool = False) -> Any:
        """
        Check the given value.

        Override this method for specific checks.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Any
            Value to check.
        recursive : bool, optional
            If True, perform the check recursively. Defaults to False.
        """
        return


# %% Constant Parameters


class Constant(ParameterBase):
    """
    Parameter that is immutable once set.

    Attributes
    ----------
    value : Any
        The immutable value of the parameter.

    Notes
    -----
    Once set, the value of a Constant parameter cannot be modified.
    """

    def __init__(self, value: Any, *args: Any, **kwargs: Any) -> None:
        """
        Initialize a Constant instance.

        Parameters
        ----------
        value : Any
            Constant value for the parameter.
        *args : Any
            Variable length argument list.
        **kwargs : Any
            Arbitrary keyword arguments.
        """
        super().__init__(*args, default=value, **kwargs)

    def __set__(self, instance: Any, value: Any) -> None:
        """
        Disallow modification of the value of a Constant parameter.

        Raises
        ------
        ValueError
            If trying to modify the constant parameter.
        """
        raise ValueError("Cannot modify constant parameter.")


class Switch(ParameterBase):
    """
    Parameter that can be set to one of several predefined options.

    Attributes
    ----------
    valid : list
        List of valid options for the parameter.

    Notes
    -----
    Assign a value to this parameter from the `valid` list.
    """

    def __init__(self, *args: Any, valid: list, **kwargs: Any) -> None:
        """
        Initialize a Switch instance.

        Parameters
        ----------
        *args : Any
            Variable length argument list.
        valid : list
            List of valid options for the parameter.
        **kwargs : Any
            Arbitrary keyword arguments.

        Raises
        ------
        TypeError
            If `valid` is not a list.
        ValueError
            If the `valid` list is empty.
        """
        if not isinstance(valid, list):
            raise TypeError("Expected a list for valid entries")
        if not valid:
            raise ValueError("The valid options list cannot be empty.")
        self.valid = valid

        super().__init__(*args, **kwargs)

    def _check(self, instance: Any, value: Any, recursive: bool = False) -> None:
        """
        Verify if the value belongs to the valid options.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Any
            The value to check.
        recursive : bool, optional
            If True, perform the check recursively. Defaults to False.

        Raises
        ------
        ValueError
            If the value isn't one of the valid options.
        """
        if value not in self.valid:
            raise ValueError(f"Value must be one of {self.valid}")

        if recursive:
            super()._check(instance, value, recursive)


# %% Typed Parameters


class Typed(ParameterBase):
    """
    Mixin for parameters constrained to a specific type.

    `Typed` extends the base `Parameter` class with type constraints. When instantiated
    with a specific type (`ty`), it ensures values are of that type or can be cast
    to that type. If `ty` is not specified during instantiation and is not predefined
    by a subclass, an error is raised.

    Attributes
    ----------
    ty : type
        Desired type for the parameter. Defaults to accepting any type. Subclasses
        can directly set this attribute.

    Methods
    -------
    cast_value(value) -> Any:
        Attempts to cast the value to the target type. By default, returns the value
        as is. Subclasses can override for specific casting behavior.
    _prepare(instance, value, recursive=False) -> Any:
        Prepares and optionally type-casts the value before checking its type. This
        method uses `cast_value` to attempt to cast the value to the required type.
    _check(instance, value, recursive=False) -> bool:
        Validates if the value matches the desired type (`ty`). Raises a TypeError if
        validation fails.

    Notes
    -----
    - If `ty` is specified during instantiation, any value assigned to this parameter
      undergoes validation against this type.
    - Override `cast_value` in subclasses for custom casting logic.
    - Assigning `None` to the parameter removes its current value from the instance.
    - An error is raised during instantiation if `ty` is neither provided nor predefined
      by a subclass.

    See Also
    --------
    Parameter
    Bool
    Integer
    Tuple
    Float
    String
    Dictionary
    """

    def __init__(self, *args: Any, ty: Optional[type] = None, **kwargs: Any) -> None:
        """
        Initialize a Typed instance.

        Parameters
        ----------
        *args : Any
            Variable length argument list.
        ty : type, optional
            The desired type for the parameter.
        **kwargs : Any
            Arbitrary keyword arguments.
        """
        if ty is not None:
            self.ty = ty
        elif not hasattr(self, "ty"):
            raise ValueError("Type must be provided either in a subclass or during instantiation.")

        super().__init__(*args, **kwargs)

    def cast_value(self, value: Any) -> Any:
        """
        Cast the type of the given value.

        This method is a placeholder. Override it if type-casting is necessary.

        Parameters
        ----------
        value : Any
            Value to cast.

        Returns
        -------
        Any
            The unmodified value.
        """
        return value

    def _prepare(self, instance: Any, value: Any, recursive: bool = False) -> Any:
        """
        Prepare and optionally type-cast the value before type validation.

        This method is called before `_check` to provide an opportunity to
        process or type-cast the value. By default, it uses the `cast_value`
        method to attempt casting the value to the desired type.

        Parameters
        ----------
        instance : Any
            The instance to which this descriptor belongs.
        value : Any
            The value to be prepared or type-cast.
        recursive : bool, optional
            If True, the preparation is done recursively by calling the parent's
            _prepare method. This can be useful if `Typed` is combined with other
            descriptor classes that might also need to process the value.
            Defaults to False.

        Returns
        -------
        Any
            The potentially type-cast value, ready for type validation.
        """
        value = self.cast_value(value)

        if recursive:
            value = super()._prepare(instance, value, recursive)

        return value

    def _check(self, instance: Any, value: Any, recursive: bool = False) -> None:
        """
        Validate the value's type against `ty`.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Any
            Value to check.
        recursive : bool, optional
            If True, perform the check recursively. Defaults to False.

        Raises
        ------
        TypeError
            If the value's type doesn't match `ty`.
        """
        if not isinstance(value, self.ty):
            raise TypeError(f"Expected type {self.ty}, got {type(value)}")

        if recursive:
            super()._check(instance, value, recursive)


class Bool(Typed):
    """
    Parameter descriptor constrained to boolean values.

    Notes
    -----
    This class also supports casting integers 0 and 1 to their boolean equivalents.
    """

    ty = bool

    def cast_value(self, value: Any) -> Union[bool, Any]:
        """
        Convert integers 0 and 1 to their respective boolean values.

        Parameters
        ----------
        value : Any
            Value to be cast.

        Returns
        -------
        Union[bool, Any]
            Boolean equivalent if value is 0 or 1; otherwise, the original value.
        """
        if isinstance(value, (int, np.bool_)) and value in [0, 1]:
            value = bool(value)
        return value


class Integer(Typed):
    """Parameter descriptor constrained to integer values."""

    ty = int


class Float(Typed):
    """
    Parameter descriptor constrained to float values.

    Notes
    -----
    This class also supports casting integers and numpy numbers to floats.
    """

    ty = float

    def cast_value(self, value: Any) -> Union[float, Any]:
        """
        Convert integers and numpy numbers to float.

        Parameters
        ----------
        value : Any
            Value to be cast.

        Returns
        -------
        Union[float, Any]
            Float equivalent if value is an integer or numpy number;
            otherwise, the original value.
        """
        try:
            value = float(np.array(value).squeeze())
        except ValueError:
            raise TypeError("Cannot cast value to float.")

        return value


class String(Typed):
    """Parameter descriptor constrained to string values."""

    ty = str


class Tuple(Typed):
    """Parameter descriptor constrained to tuple values."""

    ty = tuple


class List(Typed):
    """Parameter descriptor constrained to list values."""

    ty = list


class Dictionary(Typed):
    """Parameter descriptor constrained to dictionary (`dict`) values."""

    ty = dict


class NdArray(Typed):
    """
    Parameter descriptor constrained to np.ndarray values.

    Notes
    -----
    The `cast_value` method automatically converts lists to numpy arrays and
    wraps scalars (int or float) into single-element numpy arrays.
    """

    ty = np.ndarray

    def cast_value(self, value: Any) -> Any | np.ndarray:
        """
        Cast lists or scalars (int or float) to numpy arrays.

        Parameters
        ----------
        value : Any
            The value to be casted.

        Returns
        -------
        np.ndarray or Any
            If the value is a list or scalar (int or float),
            it returns its numpy array equivalent.
            Otherwise, it returns the value unchanged.
        """
        if isinstance(value, list):
            value = np.array(value)
        elif isinstance(value, (int, float)):
            value = np.array((value,))

        return value


class Callable(ParameterBase):
    """
    Parameter descriptor constrained to callable objects.

    Designed to ensure a given parameter is callable. This is distinct from using a type
    constraint since built-in functions (e.g., those implemented in C) won't be captured
    by `types.FunctionTypes`.

    Examples
    --------
    Here's how you might use the `Callable` class:

    >>> class MyModel:
    ...     func = Callable()
    >>> model = MyModel()
    >>> model.func = print  # This is fine as print is callable
    >>> model.func = "not_callable"  # This will raise a TypeError

    See Also
    --------
    Parameter
    Typed
    """

    def _check(self, instance: Any, value: Any, recursive: bool = False) -> None:
        """
        Check if the given value is callable.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Any
            Value to check.
        recursive : bool, optional
            If True, perform the check recursively. Defaults to False.

        Raises
        ------
        TypeError
            If the value is not callable.
        """
        if not callable(value):
            raise TypeError("Expected object to be callable")

        if recursive:
            super()._check(instance, value, recursive)


# Also check dtype
class TypedList(List, Typed):
    """Parameter descriptor for lists with elements of e specific dtype."""

    ty = list

    def __init__(self, *args: Any, dtype: Optional[type] = None, **kwargs: Any) -> None:
        """
        Initialize a Typed instance.

        Parameters
        ----------
        *args : Any
            Variable length argument list.
        ty : type, optional
            The desired type for the parameter.
        **kwargs : Any
            Arbitrary keyword arguments.
        """
        if dtype is not None:
            self.dtype = dtype
        elif not hasattr(self, "dtype"):
            raise ValueError("dtype must be provided either in a subclass or during instantiation.")

        super().__init__(*args, **kwargs)

    def _prepare(self, instance: Any, value: Any, recursive: bool = False) -> Any:
        """
        Prepare and optionally cast the array dtype.

        This method is called before `_check` to provide an opportunity to
        process or type-cast the value. By default, it uses the `cast_value`
        method to attempt casting the value to the desired type.

        Parameters
        ----------
        instance : Any
            The instance to which this descriptor belongs.
        value : Any
            The value to be prepared or type-cast.
        recursive : bool, optional
            If True, the preparation is done recursively by calling the parent's
            _prepare method. This can be useful if `Typed` is combined with other
            descriptor classes that might also need to process the value.
            Defaults to False.

        Returns
        -------
        Any
            The potentially type-cast value, ready for type validation.
        """
        try:
            value = np.array(value, dtype=self.dtype).tolist()
        except ValueError:
            raise ValueError(f"could not convert elements to {self.dtype}")

        if recursive:
            value = super()._prepare(instance, value, recursive)

        return value

    def check_dtype(self, value: Any) -> None:
        """
        Validate if the dtype of values is correct.

        Parameters
        ----------
        value : Any
            Value to check against the range.

        Raises
        ------
        TypeError
            If the dtype is not correct.
        """
        value_array = np.array(value)

        if value_array.dtype != self.dtype:
            raise ValueError(f"Value entries must be of type {self.dtype}")

    def _check(self, instance: Any, value: Any, recursive: bool = False) -> None:
        """
        Validate the value against the range.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Any
            Value to check.
        recursive : bool, optional
            If True, perform the check recursively. Defaults to False.
        """
        self.check_dtype(value)

        if recursive:
            super()._check(instance, value, recursive)


class IntegerList(TypedList):
    """List of integers."""

    dtype = int


class FloatList(TypedList):
    """List of floats."""

    dtype = float


# %% Ranged Parameters


class Ranged(ParameterBase):
    """
    Descriptor for parameters within specified bounds.

    Allows setting values constrained by provided lower and upper bounds. The actual comparisons
    against the bounds can be customized using the `lb_op` and `ub_op` comparison functions.

    Attributes
    ----------
    lb : numeric
        The lower bound of the parameter. Default is negative infinity.
    lb_op : callable
        A callable that defines the comparison operation against the lower bound.
        Default is less than (<).
    ub : numeric
        The upper bound of the parameter. Default is positive infinity.
    ub_op : callable
        A callable that defines the comparison operation against the upper bound.
        Default is greater than (>).

    Examples
    --------
    Constraining a parameter between 0 and 10:

    >>> class MyClass:
    ...     value = Ranged(lb=0, ub=10)
    >>> obj = MyClass()
    >>> obj.value = 5  # This is valid
    >>> obj.value = -5  # Raises an error

    Notes
    -----
    - By default, values are checked to be strictly within the bounds (exclusive).
      To change this behavior, use the `lb_op` and `ub_op` parameters.
    - The `check_range` method can be overridden for custom range validation logic,
      especially if the data structure isn't a simple scalar value
      (e.g. for np.ndarrays).

    See Also
    --------
    ParameterBase
    """

    def __init__(
        self,
        *args: Any,
        lb: float = -math.inf,
        lb_op: tp.Callable = operator.lt,
        ub: float = math.inf,
        ub_op: tp.Callable = operator.gt,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the Ranged descriptor.

        Parameters
        ----------
        *args :
            Parameters for Parameter Base.
        lb : numeric, optional
            Lower bound. Defaults to negative infinity.
        lb_op : callable, optional
            Comparison for the lower bound. Defaults to less than.
        ub : numeric, optional
            Upper bound. Defaults to positive infinity.
        ub_op : callable, optional
            Comparison for the upper bound. Defaults to greater than.
        **kwargs: Optional
            Additional Parameters for ParameterBase.

        """
        self.lb = lb
        self.lb_op = lb_op
        self.ub = ub
        self.ub_op = ub_op

        super().__init__(*args, **kwargs)

    def check_range(self, value: Any) -> None:
        """
        Validate if the value is within the defined range.

        Override this method if other methods for checking ranges are required
        (e.g. for np.ndarrays).

        Parameters
        ----------
        value : Any
            Value to check against the range.

        Raises
        ------
        ValueError
            If the value is outside the specified bounds.
        """
        if self.lb_op(value, self.lb):
            raise ValueError(f"Value {value} is below the lower bound of {self.lb}")
        elif self.ub_op(value, self.ub):
            raise ValueError(f"Value {value} is above the upper bound of {self.ub}")

    def _check(self, instance: Any, value: Any, recursive: bool = False) -> None:
        """
        Validate the value against the range.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Any
            Value to check.
        recursive : bool, optional
            If True, perform the check recursively. Defaults to False.
        """
        self.check_range(value)

        if recursive:
            super()._check(instance, value, recursive)


# Ranged scalar parameters
class RangedInteger(Integer, Ranged):
    """Parameter descriptor for integers parameters constrained within bounds."""

    pass


class RangedFloat(Float, Ranged):
    """Parameter descriptor for float parameters constrained within bounds."""

    pass


# Ranged list / array parameters
class RangedArray(Ranged):
    """
    Parameter descriptor for arrays with elements constrained within some bounds.

    This class extends the Ranged descriptor to support array-like structures
    (lists, numpy arrays, etc.). Each element in the array is individually checked
    against the specified bounds.

    Examples
    --------
    Constraining elements of an array parameter between 0 and 10:

    >>> class MyClass:
    ...     values = RangedArray(lb=0, ub=10)
    >>> obj = MyClass()
    >>> obj.values = [5, 7, 2]  # This is valid
    >>> obj.values = [5, -1, 2]  # Raises an error indicating the second element is
        below the lower bound.

    Notes
    -----
    - The class uses numpy for efficient element-wise comparison.
    - In case of out-of-bound values, the raised exception specifies the index/indices
      of such values.

    See Also
    --------
    Ranged
    """

    def check_range(self, value: npt.ArrayLike) -> None:
        """
        Check each element of an array-like structure against specified bounds.

        Parameters
        ----------
        value : array-like
            The array whose elements need to be checked against the range.

        Raises
        ------
        ValueError
            If any element(s) of the array are outside the specified bounds. The raised exception
            indicates the index/indices of out-of-bound values.
        """
        value_array = np.array(value)

        if np.any(self.lb_op(value_array, self.lb)):
            idx = np.where(self.lb_op(value_array, self.lb))[0]
            raise ValueError(
                f"Element(s) at index/indices {idx} below the lower bound of {self.lb}"
            )

        elif np.any(self.ub_op(value_array, self.ub)):
            idx = np.where(self.ub_op(value_array, self.ub))[0]
            raise ValueError(
                f"Element(s) at index/indices {idx} above the upper bound of {self.ub}"
            )


class RangedList(List, RangedArray):
    """Parameter descriptor for lists with elements constrained within bounds."""

    pass


class RangedNdArray(NdArray, RangedArray):
    """Parameter descriptor for numpy arrays with elements constrained within bounds."""

    pass


# Unsigned scalar parameters
class Unsigned(Ranged):
    """Parameter descriptor for non-negative parameters."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize unsigned parameter."""
        super().__init__(*args, lb=0, lb_op=operator.lt, **kwargs)


class UnsignedInteger(Integer, Unsigned):
    """Parameter descriptor for unsigned integer parameters."""

    pass


class UnsignedFloat(Float, Unsigned):
    """Parameter descriptor for unsigned floating-point parameters."""

    pass


# Unsigned list / array parameters
class UnsignedArray(Unsigned, RangedArray):
    """Parameter descriptor for arrays with non-negative elements."""

    pass


class UnsignedList(List, UnsignedArray):
    """Parameter descriptor for lists with non-negative elements."""

    pass


class UnsignedNdArray(NdArray, UnsignedArray):
    """Parameter descriptor for numpy arrays with non-negative elements."""

    pass


# %% Sized Parameters


class Sized(ParameterBase):
    """
    Descriptor for parameters with size that potentially depends on instance attributes.

    Attributes
    ----------
    size : tuple
        Expected size or dimensions of the parameter. Individual elements can be
        either integers or strings indicating other instance parameters that influence
        the size.

    Methods
    -------
    is_independent : bool
        Determines whether the size is independent of other parameters.
    get_size(value) -> int
        Calculates the size of the given value. Override for custom behavior.
    get_expected_size(instance) -> int
        Computes the expected size based on the instance's other attributes.
    check_size(instance, value)
        Validates that the provided value's size matches the expected size.
    """

    def __init__(self, *args: Any, size: Union[int, tuple], **kwargs: Any) -> None:
        """
        Initialize the Sized descriptor.

        Parameters
        ----------
        *args :
            Parameters for Parameter Base.
        size : int or tuple
            The expected size or dimensions of the parameter. Individual elements
            can be either integers or strings (indicating other instance parameters).
        **kwargs :
            Additional parameters for Parameter Base.
        """
        if not isinstance(size, tuple):
            size = (size,)

        self.size = size

        super().__init__(*args, **kwargs)

    @property
    def is_independent(self) -> bool:
        """
        Determine whether the size is independent of other parameters.

        Returns
        -------
        bool
            True if the size is independent,
            False if it depends on other instance attributes.
        """
        flag = True
        if any(isinstance(i, str) for i in self.size):
            flag = False
        return flag

    def get_size(self, value: Any) -> int | tuple[int, ...]:
        """
        Determine the size of the provided value.

        Parameters
        ----------
        value : Any
            The value for which the size needs to be calculated.

        Returns
        -------
        int | tuple[int, ...]
            Size of the value.
        """
        return len(value)

    def get_expected_size(self, instance: Any) -> int:
        """
        Compute the expected size based on the instance's other attributes.

        Parameters
        ----------
        instance : object
            The instance whose attributes determine the expected size.

        Returns
        -------
        int
            Computed expected size based on instance attributes.

        Raises
        ------
        ValueError
            If an attribute, on which the size depends, is not set.
        """
        if not self.is_independent and instance is None:
            raise ValueError("Parameter is not independent, need instance get expected size!")

        size = []
        for i in self.size:
            if isinstance(i, int):
                size.append(i)
                continue

            i_value = getattr(instance, i)
            if i_value is None:
                raise ValueError(f"Value for {i} not set.")

            size.append(i_value)

        return np.prod(size)

    def check_size(self, instance: Any, value: Any) -> None:
        """
        Validate that the provided value's size matches the expected size.

        Parameters
        ----------
        instance : object
            The instance associated with the parameter.
        value : Any
            The value whose size needs to be validated.

        Raises
        ------
        ValueError
            If the value's size does not match the expected size.
        """
        size = self.get_size(value)
        try:
            expected_size = self.get_expected_size(instance)
        except ValueError:
            if np.array(value).squeeze().ndim == 0:
                return
            raise

        if size != expected_size:
            raise ValueError(f"Expected size {expected_size}")

    def _prepare(self, instance: Any, value: Any, recursive: bool = False) -> Any:
        """
        Prepare the value by typecasting and adjusting its size.

        This method handles typecasting of the provided value to the expected type
        (`self.ty`) and potentially adjusts its size. If the size of the parameter
        depends on other instance attributes, it will retrieve that expected size and
        adjust the value accordingly.

        Parameters
        ----------
        instance : object
            The instance associated with the parameter.
        value : Any
            The value to be prepared.
        recursive : bool, optional
            If True, recursively prepares the value using the superclass's method.
            Defaults to False.

        Returns
        -------
        Any
            The prepared value, which has been potentially typecasted and resized.

        Raises
        ------
        ValueError
            If the value cannot be cast to the expected type.
        """
        if not isinstance(value, self.ty):
            try:
                expected_size = self.get_expected_size(instance)
            except ValueError:
                expected_size = None

            if isinstance(value, (int, float)):
                value = self.ty((value,))
            else:
                raise ValueError("Cannot cast value from given value.")

            if expected_size == 0:
                value = None
            elif expected_size is not None:
                value = np.prod(expected_size) * value

        if recursive:
            value = super()._prepare(instance, value, recursive)

        return value

    def _check(self, instance: Any, value: Any, recursive: bool = False) -> None:
        """
        Validate the provided value's size.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Any
            Value to check.
        recursive : bool, optional
            If True, perform the check recursively. Defaults to False.
        """
        self.check_size(instance, value)

        if recursive:
            super()._check(instance, value, recursive)


class SizedList(List, Sized):
    """Descriptor for lists whose size may depend on other instance attributes."""

    pass


class SizedTuple(Tuple, Sized):
    """Descriptor for tuples whose size may depend on other instance attributes."""

    pass


class SizedNdArray(NdArray, Sized):
    """Descriptor for NumPy arrays whose size may depend on other instance attributes."""

    def get_size(self, value: np.ndarray) -> tuple[int, ...]:
        """
        Determine the size of the provided value.

        Parameters
        ----------
        value : Any
            The value for which the size needs to be determined.

        Returns
        -------
        tuple[int, ...]
            Size of the value.
        """
        return value.shape

    def get_expected_size(self, instance: Optional[Any]) -> tuple:
        """
        Calculate the expected size of a numpy array based on the instance's other attributes.

        Returns
        -------
        tuple
            The expected shape of the array.
        """
        if not self.is_independent and instance is None:
            raise ValueError("Parameter is not independent, need instance!")

        expected_size = []
        for i in self.size:
            if isinstance(i, int):
                expected_size += [i]
                continue

            dim = getattr(instance, i)
            if dim is None:
                raise ValueError(f"Value for {i} not set.")

            if isinstance(dim, np.ndarray):
                dim = list(dim.shape)
            elif isinstance(dim, int):
                dim = [dim]

            expected_size += dim

        expected_size = tuple(expected_size)

        return expected_size

    def _prepare(
        self,
        instance: object,
        value: Union[int, float, npt.ArrayLike],
        recursive: bool = False,
    ) -> np.ndarray:
        """
        Prepare the value for a NumPy array, ensuring it has the expected shape.

        If the value is a scalar (int or float), this method will transform it into a NumPy
        array with the expected shape, filled with the scalar value. If the value is
        already an array, this method ensures its shape matches the expected shape.

        Parameters
        ----------
        instance : object
            The instance associated with the parameter.
        value : Any
            The value to be prepared.
        recursive : bool, optional
            If True, recursively prepares the value using the superclass's method.
            Defaults to False.

        Returns
        -------
        np.ndarray
            The prepared NumPy array with the correct shape.

        Raises
        ------
        ValueError
            If the value cannot be cast to a NumPy array with the expected shape.
        """
        if np.array(value).squeeze().ndim == 0:
            try:
                expected_size = self.get_expected_size(instance)
            except ValueError:
                expected_size = None

            if expected_size is not None:
                value = value * np.ones(expected_size)
        else:
            try:
                value_array = np.array(value)
            except ValueError:
                raise TypeError("Cannot cast value from given value.")
            self.check_size(instance, value_array)

        if recursive:
            value = super()._prepare(instance, value, recursive)

        return value


class SizedRangedList(RangedList, SizedList):
    """Descriptor for ranged lists whose size depends on other instance attributes."""

    pass


class SizedUnsignedList(UnsignedList, SizedList):
    """Descriptor for unsigned lists whose size depends on other instance attributes."""

    pass


class SizedUnsignedNdArray(UnsignedNdArray, SizedNdArray):
    """Descriptor for unsigned numpy arrays whose size depends on other instance attributes."""

    pass


# Also check dtype
class SizedFloatList(FloatList, SizedList):
    """Descriptor for lists of floats whose size depends on other instance attributes."""

    pass


class SizedRangedIntegerList(RangedList, IntegerList, SizedList):
    """Descriptor for ranged lists of integers whose size depends on other instance attributes."""

    pass


class SizedUnsignedIntegerList(UnsignedList, IntegerList, SizedList):
    """Descriptor for unsigned lists of integers whose size depends on other instance attributes."""

    pass


# %% Dimensionalized Parameters


class DimensionalizedArray(NdArray):
    """
    Parameter descriptor constrained to np.arrays with a specific dimensionality.

    The descriptor ensures that the ndarray assigned matches the specified
    number of dimensions (n_dim).

    Attributes
    ----------
    n_dim : int
        The number of dimensions the array must have.

    Examples
    --------
    To create a descriptor for 2-dimensional arrays:

    >>> class MyClass:
    ...     arr = DimensionalizedArray(n_dim=2)
    >>> obj = MyClass()
    >>> obj.arr = np.array([[1, 2], [3, 4]])  # This is valid
    >>> obj.arr = np.array([1, 2, 3, 4])  # Raises a ValueError

    Notes
    -----
    The n_dim attribute can be set during initialization.
    """

    n_dim = None

    def __init__(self, n_dim: Optional[int] = None, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the DimensionalizedArray descriptor.

        Parameters
        ----------
        n_dim : int, optional
            The number of dimensions the array must have. If not specified,
            it must be set before usage.
        *args : optional
            Parameters for NdArray.
        **kwargs : optional
            Additional Parameters for NdArray.

        Raises
        ------
        ValueError
            If n_dim is not an integer.
        """
        super().__init__(*args, **kwargs)

        if n_dim is not None:
            if not isinstance(n_dim, int):
                raise ValueError("Dimensionality (n_dim) must be an integer.")
            self.n_dim = n_dim

        # Ensure the dimension is set and valid
        if self.n_dim is None:
            raise ValueError("Dimensionality (n_dim) must be set during initialization.")

    def _check(self, instance: Any, value: npt.ArrayLike, recursive: bool = False) -> None:
        """
        Check the dimensionality of the given value.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        value : Any
            Value to check.
        recursive : bool, optional
            If True, perform the check recursively. Defaults to False.

        Raises
        ------
        ValueError
            If the number of dimensions of the ndarray doesn't match n_dim.
        """
        if self.n_dim != np.array(value).ndim:
            raise ValueError(f"Expected {self.n_dim} dimensions, got {np.array(value).ndim}.")

        if recursive:
            super()._check(instance, value, recursive)


class Vector(DimensionalizedArray):
    """
    Parameter descriptor for one-dimensional numpy arrays (vectors).

    Attributes
    ----------
    n_dim : int
        Dimensionality of the numpy array, set to 1 for vectors.

    Examples
    --------
    >>> class MyModel:
    ...     coordinates = Vector()
    >>> model.coordinates = np.array([1, 2, 3])  # Valid
    >>> model.coordinates = np.array([[1, 2], [3, 4]])  # Raises ValueError

    See Also
    --------
    Dimensionalized
    NdArray
    """

    n_dim = 1


class Matrix(DimensionalizedArray):
    """
    Parameter descriptor for two-dimensional numpy arrays (matrices).

    This descriptor ensures that the ndarray assigned is two-dimensional.

    Attributes
    ----------
    n_dim : int
        Dimensionality of the numpy array, set to 2 for matrices.

    Examples
    --------
    >>> class MyModel:
    ...     data = Matrix()
    >>> model = MyModel()
    >>> model.data = np.array([[1, 2], [3, 4]])  # Valid
    >>> model.data = np.array([1, 2, 3, 4])  # Raises ValueError

    See Also
    --------
    DimensionalizedArray
    NdArray
    """

    n_dim = 2


# %% Polynomial Parameters


class NdPolynomial(SizedNdArray):
    """
    Dependently sized polynomial for n entries.

    This descriptor represents a polynomial whose size or dimensions may depend on other
    instance attributes. The polynomial can also be thought of as a 2D array, where each
    row represents a polynomial of a certain degree.

    Important: Use [entries x n_coeff] for dependencies.

    Parameters
    ----------
    n_entries : int, optional
        Number of polynomials or rows. Default is None.
    n_coeff : int, optional
        Number of coefficients for each polynomial or columns. Default is None.

    Attributes
    ----------
    size : tuple
        The shape of the polynomial array, determined from n_entries and n_coeff.

    Notes
    -----
    Currently, NdPolynomial is implemented as `SizedNdArray`.
    Consequently, no default values can be set since their size would depend on the
    dependent variables.
    In theory, this could be split into `NdPolynomial` and `SizedNdPolynomial`,
    but there is currently no use for this distinction.

    Methods
    -------
    fill_values(dims, value) -> np.ndarray:
        Fills values to generate the polynomial matrix of the desired size.
    _prepare(instance, value, recursive=False) -> np.ndarray:
        Prepare the given polynomial matrix s.t. it adheres to the expected size.
    """

    def __init__(
        self,
        *args: Any,
        n_entries: Optional[int] = None,
        n_coeff: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize an NdPolynomial descriptor with specific entries and coefficients.

        Parameters
        ----------
        n_entries : int, optional
            Set the number of polynomials or rows. If not provided, it must be inferred
            from 'size'.
        n_coeff : int, optional
            Define the number of coefficients for each polynomial or columns. If not
            provided, it must be inferred from 'size'.
        *args :
            Variable length argument list.
        **kwargs :
            Arbitrary keyword arguments, can include 'size' which denotes the shape of
            the polynomial array.

        Raises
        ------
        ValueError
            If both 'n_entries' and 'n_coeff' are missing and 'size' is not provided or
            is incomplete.
            If there's a mismatch or redundancy in provided dimensions.

        Notes
        -----
        The shape of the polynomial array is determined from 'n_entries' and 'n_coeff'
        or from the 'size' keyword argument.
        """
        if "default" in kwargs and kwargs["default"] != 0:
            raise ValueError("Default value for NdPolynomial must always be 0.")

        try:
            size = kwargs["size"]
            if not isinstance(size, tuple):
                size = (size,)
        except KeyError:
            size = tuple()

        if n_entries is None and n_coeff is None and len(size) < 2:
            raise ValueError("Missing value for n_coeff for shape")

        if n_entries is not None:
            if len(size) > 1:
                raise ValueError("Found duplicate n_entries for shape")
            _n_entries = n_entries
            if n_coeff is not None:
                if len(size) > 0:
                    raise ValueError("Found duplicate n_entries for shape")
        else:
            _n_entries = size[0]

        if n_coeff is not None:
            if len(size) > 1:
                raise ValueError("Found duplicate n_entries for shape")
            _n_coeff = n_coeff
            if n_entries is not None:
                if len(size) > 0:
                    raise ValueError("Found duplicate n_entries for shape")
        else:
            if n_entries is not None:
                _n_coeff = size[0]
            else:
                _n_coeff = size[1]

        size = (_n_entries, _n_coeff)

        kwargs["size"] = size

        super().__init__(*args, **kwargs)

    def fill_values(
        self, dims: tuple[int], value: Union[int, float, np.ndarray, list]
    ) -> np.ndarray:
        """
        Fill values to generate the polynomial matrix of the desired size.

        Parameters
        ----------
        dims : tuple of int
            Dimensions (n_entries, n_coeff) of the polynomial matrix.
        value : Union[int, float, np.ndarray, list]
            Value(s) to be filled in the polynomial matrix.

        Returns
        -------
        np.ndarray
            Polynomial matrix of the desired size filled with given values.

        Raises
        ------
        ValueError
            If there's a mismatch between the provided values and expected dimensions.
        """
        if len(dims) == 1:
            n_entries = 1
            n_coeff = dims[0]
            single_entry = True
            value = np.array(value, ndmin=2)
        else:
            n_entries = dims[0]
            n_coeff = dims[1]
            single_entry = False

        if single_entry and n_entries > 1:
            raise ValueError("Can only set single entry if n_entries == 1.")

        if isinstance(value, np.ndarray) and value.size == 1:
            value = float(value.squeeze())

        if isinstance(value, (int, float)):
            value = n_entries * [value]
        elif isinstance(value, np.ndarray):
            value = value.tolist()

        if len(value) != n_entries:
            raise ValueError("Number of entries does not match")

        _value = np.zeros((n_entries, n_coeff))

        for i, v in enumerate(value):
            if isinstance(v, (int, float, np.number)):
                v = [v]
            if isinstance(v, (list, tuple)):
                missing = n_coeff - len(v)
                v += missing * (0,)
            _value[i, :] = np.array(v)

        if single_entry:
            _value = _value[0]

        return _value

    def _prepare(
        self,
        instance: object,
        value: Union[int, float, np.ndarray, list],
        recursive: bool = False,
    ) -> np.ndarray:
        """
        Prepare the given polynomial matrix s.t. it adheres to the expected size.

        Parameters
        ----------
        instance : object
            The instance whose attributes might influence the size.
        value : Union[int, float, np.ndarray, list]
            Value(s) to be filled in the polynomial matrix.
        recursive : bool, optional
            If True, perform the operation recursively. Defaults to False.

        Returns
        -------
        np.ndarray
            Polynomial matrix of the desired size prepared as per requirements.
        """
        if instance is not None:
            dims = self.get_expected_size(instance)
            value = self.fill_values(dims, value)

        if recursive:
            value = super()._prepare(instance, value, recursive)

        return value


class Polynomial(NdPolynomial):
    """
    Represent a single polynomial using coefficients.

    This class serves as a simplified version of NdPolynomial, specifically tailored for
    single polynomials. It is always defined by its coefficients, removing the need for
    an 'n_entries' parameter.

    Use this class when only a single polynomial representation is required.

    Attributes
    ----------
    size : tuple
        Size of the coefficients for the polynomial. Derived from NdPolynomial but omits
        'n_entries'.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize a Polynomial descriptor with a specific coefficient length.

        Parameters
        ----------
        *args :
            Variable length argument list.
        **kwargs :
            Arbitrary keyword arguments. Typically used for any parameters inherited
            from NdPolynomial, except 'n_entries'.

        Notes
        -----
        The Polynomial descriptor is a special case of NdPolynomial with 'n_entries'
        always set to 1. Therefore, it's only defined by its coefficients.
        """
        super().__init__(*args, n_entries=1, **kwargs)
        self.size = self.size[1:]


# %% Modulated Parameters


class DependentlyModulated(Sized):
    """
    Mixin for checking parameter shapes based on other instance attributes.

    This mixin ensures that the size of a parameter is modulo an expected size. If this
    condition is not met, a ValueError is raised.
    """

    def check_mod_value(self, instance: Any, value: Any) -> None:
        """
        Check if the size of the parameter modulo its expected size is zero.

        By default, this method checks the modulo condition, but can be overridden
        by subclasses to incorporate custom behaviors.

        Parameters
        ----------
        instance : Any
            The instance associated with the parameter.
        value : Any
            The value whose size needs to be validated.

        Raises
        ------
        ValueError
            If the modulo condition of the size does not meet the expected criteria.
        """
        size = self.get_size(value)
        expected_size = self.get_expected_size(instance)

        size %= expected_size

        if size != 0:
            raise ValueError(
                f"The size of the value modulo the expected size is not zero. "
                f"Size: {size}, Expected Size: {expected_size}"
            )

    check_size = check_mod_value


class DependentlyModulatedUnsignedList(UnsignedList, SizedList, DependentlyModulated):
    """List of unsigned values whose size is dependent on other attributes."""

    pass



"""
Pseudo-transient continuation for solving nonlinear equation systems.

The main method is ptc, a pseudo-transient continuation method with switched evolution
relaxation.

This code was written by Samuel Leweke (University of Cologne) in 2020.
"""  # noqa

__author__ = "Samuel Leweke"
__contact__ = "leweke@math.uni-koeln.de"
__copyright__ = "Copyright 2020, University of Cologne"

import math
from typing import Callable, Optional

import numpy as np
import scipy.linalg


def scaled_norm2(x: np.ndarray, scale: np.ndarray) -> float:
    """Calculate the scaled Euclidean (L2) norm."""
    return math.sqrt(np.sum(np.square(x / scale)) / len(x))


def norm2(x: np.ndarray) -> float:
    """Calculate the karthesian norm."""
    return math.sqrt(np.sum(np.square(x)) / len(x))


def jacrow_scale(jacMat: np.ndarray, scale: np.ndarray) -> np.ndarray:
    """Scale jacMat with scale."""
    return jacMat / scale[:, None]


def ptc(
    x: np.ndarray,
    f: Callable[[np.ndarray], np.ndarray],
    jacF: Callable[[np.ndarray], np.ndarray],
    tau: float,
    tol: float,
    scale: Optional[np.ndarray] = None,
    maxIter: int = 50,
    maxNonMonotone: int = 5,
    quiet: bool = True,
    variant: bool = False,
) -> tuple[int, np.ndarray, float, int]:
    r"""
    Solve a nonlinear equation system using pseudo-transient continuation.

    The nonlinear equation system f(x) = 0 is solved using pseudo-transient
    continuation (PTC), which introduce pseudo time and computes the steady
    state of the ODE dx / dt = f(x).

    After resolving the transient regime at the beginning of the time
    integration, PTC quickly increases the time step size to arrive at the
    equilibrium. It can be interpreted as a linear-implicit Euler scheme for
    the ODE, which tends to a Newton method as the time step size increases.

    The step size is adaptively determined by switched evolution relaxation
    (SER).

    See Deuflhard, Peter (2011). Newton Methods for Nonlinear Problems
    (Vol. 35). Berlin, Heidelberg: Springer.
    http://doi.org/10.1007/978-3-642-23899-4


    Parameters
    ----------
    x : list
        Initial point as numpy array.
    f : callable
        Function with signature f(x) that returns the residual.
    jacF : callable
        Function with signature jacF(x) that returns the Jacobian of f.
    tau : TYPE
        Initial pseudo time step size.
    tol : TYPE
        Target tolerance for scaled root mean square residual.
    scale : np.ndarray, optional
        (positive) diagonal scaling coefficients.
        The scaled root mean square norm is given by
        .. math::
            || x || = \sqrt{ (1/n) * \sum (x_i / v_i)^2}
        If None, no scaling is applied. The default is None.
    maxIter : int, optional
        Maximum number of iterations. The default is 50.
    maxNonMonotone : int, optional
        Maximum number of iterations with non-decreasing residual in a row.
        The default is 5.
    quiet : bool, optional
        If True, norm of step, scaled root mean square residual, and step size
        are printed on each iteration. The default is True.
    variant : bool, optional
        If True, (1/tau * I - F'(x)) * dx = F(x) is solved.
        If False, (I - tau * F'(x)) * dx = F(x) is solved.
        The default is False

    Returns
    -------
    status : int
        -2: Failed due to non-decreasing residual
        -1: Failed due to singular Jacobian
         0: Converged with passing residual test
         1: Exceeded maximum number of iterations.
    x : np.ndarray
        Solution.
    normfk : np.ndarray
        Scaled root mean square residual.
    k : int
        Number of iterations.
    """
    fxk = f(x)
    if scale is None:
        scale = np.ones(x.shape)

    normfk = scaled_norm2(fxk, scale)
    normdx = 0.0

    k = 0
    nNonMonotoneResidual = 0

    if not quiet:
        print("%4i  %12e  %12e  %7f" % (k, normdx, normfk, tau))

    while True:
        if normfk <= tol:
            return (0, x, normfk, k)

        jacMat = jacF(x)
        jacMat = jacrow_scale(jacMat, scale)

        if variant:
            # Solve scaled system
            # (1/tau * I - F'(x)) * dx = F(x), update x <- x + dx
            jacQ, jacR = scipy.linalg.qr(np.diag(1 / (tau * scale)) - jacMat)
        else:
            # Solve scaled system
            # (I - tau * F'(x)) * dx = F(x), update x <- x + tau * dx
            jacQ, jacR = scipy.linalg.qr(np.diag(1 / scale) - tau * jacMat)

        dx = fxk / scale
        try:
            dx = scipy.linalg.solve_triangular(jacR, np.dot(jacQ.T, dx))
        except scipy.linalg.LinAlgError:
            # Singular matrix
            return (-1, x, normfk, k)

        # Update current position
        if variant:
            x = x + dx
        else:
            x = x + tau * dx

        # Calculate residual at new position
        fxk = f(x)
        normfkp1 = scaled_norm2(fxk, scale)

        # Protect against ZeroDivisionError
        if normfkp1 == 0.0:
            return (0, x, normfkp1, k)

        # Update step size
        tau = tau * normfk / normfkp1

        # Check monotonicity of residual
        if normfkp1 < normfk:
            nNonMonotoneResidual = 0
        else:
            nNonMonotoneResidual = nNonMonotoneResidual + 1
            if nNonMonotoneResidual >= maxNonMonotone:
                return (-2, x, normfkp1, k)

        normfk = normfkp1
        k = k + 1

        if not quiet:
            normdx = norm2(dx)
            print("%4i  %12e  %12e  %7f" % (k, normdx, normfk, tau))

        # Check terminal condition
        if k > maxIter:
            break

    return (1, x, normfk, k)
